#
# This file is part of LiteDRAM.
#
# Copyright (c) 2018-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

"""
ECC frontend for LiteDRAM

Adds ECC support to Native ports.

Features:
- Single Error Correction.
- Double Error Detection.
- Errors injection.
- Errors reporting.

Limitations:
- Write byte enable granularity of DRAM's data-width.
"""

from migen import *

from litex.soc.interconnect.csr import *
from litex.soc.interconnect.stream import *
from litex.soc.cores.ecc import *

from litedram.common import wdata_description, rdata_description


# LiteDRAMNativePortECCW ---------------------------------------------------------------------------

class LiteDRAMNativePortECCW(Module):
    def __init__(self, data_width_from, data_width_to, burst_cycles=8):
        self.sink     = sink   = Endpoint(wdata_description(data_width_from))
        self.source   = source = Endpoint(wdata_description(data_width_to))
        self.we_error = Signal()

        # # #

        # Control (ECC encoding is combinatorial).
        self.comb += sink.connect(source, omit={"data", "we"}),

        # Data Path.
        for i in range(burst_cycles):
            ecc_width_from = data_width_from // burst_cycles
            ecc_width_to   = data_width_to   // burst_cycles

            # Encoder.
            self.submodules.encoder = encoder = ECCEncoder(ecc_width_from)
            self.comb += [
                # Input.
                encoder.i.eq(sink.data[i*ecc_width_from:(i+1)*ecc_width_from]),
                # Output.
                # We always have to store the full ECC-Word so byte enable is not supported within
                # an ECC word. If any of the byte enable is set, the full ECC-Word is written.
                If(sink.we[i*ecc_width_from//8:(i+1)*ecc_width_from//8] != 0,
                    source.we[i*ecc_width_to//8:(i+1)*ecc_width_to//8].eq(2**ecc_width_to//8-1)
                ),
                source.data[i*ecc_width_to:(i+1)*ecc_width_to].eq(encoder.o),
                # Byte enable granularity  error detection.
                If(sink.valid & (sink.we[i*ecc_width_from//8:(i+1)*ecc_width_from//8] != (2**ecc_width_from//8-1)),
                    self.we_error.eq(1)
                )
            ]

# LiteDRAMNativePortECCR ---------------------------------------------------------------------------

class LiteDRAMNativePortECCR(Module):
    def __init__(self, data_width_from, data_width_to, burst_cycles=8):
        self.sink   = sink   = Endpoint(rdata_description(data_width_to))
        self.source = source = Endpoint(rdata_description(data_width_from))
        self.enable = Signal()
        self.sec    = Signal(burst_cycles)
        self.ded    = Signal(burst_cycles)

        # # #

        # Control Path (ECC decoding is combinatorial).
        self.comb +=  sink.connect(source, omit={"data"})

        # Data Path.
        for i in range(burst_cycles):
            ecc_width_to   = data_width_to   // burst_cycles
            ecc_width_from = data_width_from // burst_cycles

            # Decoder.
            self.submodules.decoder = decoder = ECCDecoder(ecc_width_from)
            self.comb += [
                # Enable.
                decoder.enable.eq(self.enable),
                # Input.
                decoder.i.eq(sink.data[i*ecc_width_to:(i+1)*ecc_width_to]),
                # Output.
                source.data[i*ecc_width_from:(i+1)*ecc_width_from].eq(decoder.o),
                # Bitflip/Error reporting.
                If(source.valid,
                    self.sec[i].eq(decoder.sec),
                    self.ded[i].eq(decoder.ded)
                )
            ]

# LiteDRAMNativePortECC ----------------------------------------------------------------------------

class LiteDRAMNativePortECC(Module, AutoCSR):
    def __init__(self, port_from, port_to, burst_cycles=8, with_error_injection=False, with_we_error_detection=False):
        _ , n = compute_m_n(port_from.data_width//burst_cycles)
        assert port_to.data_width >= (n + 1)*burst_cycles

        self.enable     = CSRStorage(reset=1)
        self.clear      = CSR()
        self.sec_errors = CSRStatus(32)
        self.ded_errors = CSRStatus(32)
        self.sec_detected = sec_detected = Signal()
        self.ded_detected = ded_detected = Signal()
        if with_error_injection:
            self.flip = CSRStorage(8)
        if with_we_error_detection:
            self.we_errors = CSRStatus(32)

        # # #

        # Cmd --------------------------------------------------------------------------------------
        self.comb += port_from.cmd.connect(port_to.cmd)

        # Wdata (ECC) encoding) --------------------------------------------------------------------
        ecc_wdata = LiteDRAMNativePortECCW(port_from.data_width, port_to.data_width, burst_cycles)
        ecc_wdata = BufferizeEndpoints({"source": DIR_SOURCE})(ecc_wdata)
        self.submodules += ecc_wdata
        self.comb += [
            port_from.wdata.connect(ecc_wdata.sink),
            ecc_wdata.source.connect(port_to.wdata)
        ]
        if with_error_injection:
            self.comb += port_to.wdata.data[:8].eq(self.flip.storage ^ ecc_wdata.source.data[:8])

        # Rdata (ECC decoding) ---------------------------------------------------------------------
        sec = Signal()
        ded = Signal()
        ecc_rdata = LiteDRAMNativePortECCR(port_from.data_width, port_to.data_width, burst_cycles)
        ecc_rdata = BufferizeEndpoints({"source": DIR_SOURCE})(ecc_rdata)
        self.submodules += ecc_rdata
        self.comb += [
            ecc_rdata.enable.eq(self.enable.storage),
            port_to.rdata.connect(ecc_rdata.sink),
            ecc_rdata.source.connect(port_from.rdata)
        ]

        # Errors count -----------------------------------------------------------------------------
        sec_errors = self.sec_errors.status
        ded_errors = self.ded_errors.status
        self.sync += [
            If(self.clear.re,
                sec_errors.eq(0),
                ded_errors.eq(0),
                sec_detected.eq(0),
                ded_detected.eq(0),
            ).Else(
                If(sec_errors != (2**len(sec_errors) - 1),
                    If(ecc_rdata.sec != 0,
                        sec_detected.eq(1),
                        sec_errors.eq(sec_errors + 1)
                    )
                ),
                If(ded_errors != (2**len(ded_errors) - 1),
                    If(ecc_rdata.ded != 0,
                        ded_detected.eq(1),
                        ded_errors.eq(ded_errors + 1)
                    )
                )
            )
        ]
        if with_we_error_detection:
            we_errors = self.we_errors.status
            self.sync += [
                If(self.clear.re,
                    we_errors.eq(0),
                ).Else(
                    If(we_errors != (2**len(we_errors) - 1),
                        If(ecc_wdata.we_error,
                            we_errors.eq(we_errors + 1)
                        )
                    )
                )
            ]



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.soc.interconnect.csr import CSR

from litedram.phy.utils import delayed, Serializer, Deserializer, Latency
from litedram.phy.sim_utils import SimPad, SimulationPads, SimSerDesMixin
from litedram.phy.lpddr4.basephy import LPDDR4PHY, DoubleRateLPDDR4PHY


class LPDDR4SimulationPads(SimulationPads):
    def layout(self, databits=16):
        return [
            SimPad("clk", 1),
            SimPad("cke", 1),
            SimPad("odt", 1),
            SimPad("reset_n", 1),
            SimPad("cs", 1),
            SimPad("ca", 6),
            SimPad("dq", databits, io=True),
            SimPad("dqs", databits//8, io=True),
            SimPad("dmi", databits//8, io=True),
        ]


class LPDDR4SimPHY(SimSerDesMixin, LPDDR4PHY):
    """LPDDR4 simulation PHY with direct 16:1 serializers

    For simulation purpose two additional "DDR" clock domains are requires.
    """
    def __init__(self, aligned_reset_zero=False, **kwargs):
        pads = LPDDR4SimulationPads()
        self.submodules += pads
        super().__init__(pads,
            ser_latency  = Latency(sys=Serializer.LATENCY),
            des_latency  = Latency(sys=Deserializer.LATENCY),
            phytype      = "LPDDR4SimPHY",
            **kwargs)

        # fake delays (make no nsense in simulation, but sdram.c expects them)
        self.settings.read_leveling = True
        self.settings.delays = 1
        self._rdly_dq_rst = CSR()
        self._rdly_dq_inc = CSR()

        delay = lambda sig, cycles: delayed(self, sig, cycles=cycles)
        sdr    = dict(clkdiv="sys", clk="sys8x")
        sdr_90 = dict(clkdiv="sys", clk="sys8x_90")
        ddr    = dict(clkdiv="sys", clk="sys8x_ddr")
        ddr_90 = dict(clkdiv="sys", clk="sys8x_90_ddr")

        if aligned_reset_zero:
            sdr["reset_cnt"] = 0
            ddr["reset_cnt"] = 0

        # Clock is shifted 180 degrees to get rising edge in the middle of SDR signals.
        # To achieve that we send negated clock on clk (clk_p).
        self.ser(i=~self.out.clk, o=self.pads.clk, name='clk', **ddr)

        self.ser(i=self.out.cke, o=self.pads.cke, name='cke', **sdr)
        self.ser(i=self.out.odt, o=self.pads.odt, name='odt', **sdr)
        self.ser(i=self.out.reset_n, o=self.pads.reset_n, name='reset_n', **sdr)

        # Command/address
        self.ser(i=self.out.cs, o=self.pads.cs, name='cs', **sdr)
        for i in range(6):
            self.ser(i=self.out.ca[i], o=self.pads.ca[i], name=f'ca{i}', **sdr)

        # Tristate I/O (separate for simulation)
        for i in range(self.databits//8):
            self.ser(i=self.out.dmi_o[i], o=self.pads.dmi_o[i], name=f'dmi_o{i}', **ddr)
            self.des(o=self.out.dmi_i[i], i=self.pads.dmi[i],   name=f'dmi_i{i}', **ddr)
            self.ser(i=self.out.dqs_o[i], o=self.pads.dqs_o[i], name=f'dqs_o{i}', **ddr_90)
            self.des(o=self.out.dqs_i[i], i=self.pads.dqs[i],   name=f'dqs_i{i}', **ddr_90)
        for i in range(self.databits):
            self.ser(i=self.out.dq_o[i], o=self.pads.dq_o[i], name=f'dq_o{i}', **ddr)
            self.des(o=self.out.dq_i[i], i=self.pads.dq[i],   name=f'dq_i{i}', **ddr)

        # Output enable signals
        self.comb += [
            self.pads.dmi_oe.eq(delay(self.out.dmi_oe, cycles=Serializer.LATENCY)),
            self.pads.dqs_oe.eq(delay(self.out.dqs_oe, cycles=Serializer.LATENCY)),
            self.pads.dq_oe.eq(delay(self.out.dq_oe, cycles=Serializer.LATENCY)),
        ]


class DoubleRateLPDDR4SimPHY(SimSerDesMixin, DoubleRateLPDDR4PHY):
    """LPDDR4 simulation PHY basing of DoubleRateLPDDR4PHY

    `DoubleRateLPDDR4PHY` performs a single serialization step between `sys` and `sys2x`,
    so this PHY wrapper has to do the serialization between `sys2x` and `sys8x` (SDR/DDR).

    For simulation purpose two additional "DDR" clock domains are requires.
    """
    def __init__(self, aligned_reset_zero=False, **kwargs):
        pads = LPDDR4SimulationPads()
        self.submodules += pads
        super().__init__(pads,
            ser_latency  = Latency(sys2x=Serializer.LATENCY),
            des_latency  = Latency(sys2x=Deserializer.LATENCY),
            phytype      = "LPDDR4SimPHY",
            **kwargs)
        self.submodules.half_delay = ClockDomainsRenamer("sys2x")(Module())

        # fake delays (make no nsense in simulation, but sdram.c expects them)
        self.settings.read_leveling = True
        self.settings.delays = 1
        self._rdly_dq_rst = CSR()
        self._rdly_dq_inc = CSR()

        delay = lambda sig, cycles: delayed(self.half_delay, sig, cycles=cycles)

        sdr    = dict(clkdiv="sys2x", clk="sys8x")
        sdr_90 = dict(clkdiv="sys2x", clk="sys8x_90")
        ddr    = dict(clkdiv="sys2x", clk="sys8x_ddr")
        ddr_90 = dict(clkdiv="sys2x", clk="sys8x_90_ddr")

        if aligned_reset_zero:
            sdr["reset_cnt"] = 0
            ddr["reset_cnt"] = 0

        # Clock is shifted 180 degrees to get rising edge in the middle of SDR signals.
        # To achieve that we send negated clock on clk (clk_p).
        self.ser(i=~self.out.clk, o=self.pads.clk, name='clk', **ddr)

        self.ser(i=self.out.cke, o=self.pads.cke, name='cke', **sdr)
        self.ser(i=self.out.odt, o=self.pads.odt, name='odt', **sdr)
        self.ser(i=self.out.reset_n, o=self.pads.reset_n, name='reset_n', **sdr)

        # Command/address
        self.ser(i=self.out.cs, o=self.pads.cs, name='cs', **sdr)
        for i in range(6):
            self.ser(i=self.out.ca[i], o=self.pads.ca[i], name=f'ca{i}', **sdr)

        # Tristate I/O (separate for simulation)
        for i in range(self.databits//8):
            self.ser(i=self.out.dmi_o[i], o=self.pads.dmi_o[i], name=f'dmi_o{i}', **ddr)
            self.des(o=self.out.dmi_i[i], i=self.pads.dmi[i],   name=f'dmi_i{i}', **ddr)
            self.ser(i=self.out.dqs_o[i], o=self.pads.dqs_o[i], name=f'dqs_o{i}', **ddr_90)
            self.des(o=self.out.dqs_i[i], i=self.pads.dqs[i],   name=f'dqs_i{i}', **ddr_90)
        for i in range(self.databits):
            self.ser(i=self.out.dq_o[i], o=self.pads.dq_o[i], name=f'dq_o{i}', **ddr)
            self.des(o=self.out.dq_i[i], i=self.pads.dq[i],   name=f'dq_i{i}', **ddr)

        # Output enable signals
        self.comb += [
            self.pads.dmi_oe.eq(delay(self.out.dmi_oe, cycles=Serializer.LATENCY)),
            self.pads.dqs_oe.eq(delay(self.out.dqs_oe, cycles=Serializer.LATENCY)),
            self.pads.dq_oe.eq(delay(self.out.dq_oe, cycles=Serializer.LATENCY)),
        ]



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2015-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2020-2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

# SDRAM simulation PHY at DFI level tested with SDR/DDR/DDR2/LPDDR/DDR3
# TODO:
# - add multirank support.

from migen import *

from litedram.common import *
from litedram.phy.dfi import *
from litedram.modules import _speedgrade_timings, _technology_timings

from functools import reduce
from operator import or_

import struct


SDRAM_VERBOSE_OFF = 0
SDRAM_VERBOSE_STD = 1
SDRAM_VERBOSE_DBG = 2

# Bank Model ---------------------------------------------------------------------------------------

class BankModel(Module):
    def __init__(self, data_width, nrows, ncols, burst_length, nphases, we_granularity, init):
        self.activate     = Signal()
        self.activate_row = Signal(max=nrows)
        self.precharge    = Signal()

        self.write        = Signal()
        self.write_col    = Signal(max=ncols)
        self.write_data   = Signal(data_width)
        self.write_mask   = Signal(data_width//8)

        self.read         = Signal()
        self.read_col     = Signal(max=ncols)
        self.read_data    = Signal(data_width)

        # # #

        active = Signal()
        row    = Signal(max=nrows)

        self.sync += \
            If(self.precharge,
                active.eq(0),
            ).Elif(self.activate,
                active.eq(1),
                row.eq(self.activate_row)
            )

        bank_mem_len   = nrows*ncols//(burst_length*nphases)
        mem            = Memory(data_width, bank_mem_len, init=init)
        write_port     = mem.get_port(write_capable=True, we_granularity=we_granularity)
        read_port      = mem.get_port(async_read=True)
        self.specials += mem, read_port, write_port

        wraddr         = Signal(max=bank_mem_len)
        rdaddr         = Signal(max=bank_mem_len)

        self.comb += [
            wraddr.eq((row*ncols | self.write_col)[log2_int(burst_length*nphases):]),
            rdaddr.eq((row*ncols | self.read_col)[log2_int(burst_length*nphases):]),
        ]

        self.comb += [
            If(active,
                write_port.adr.eq(wraddr),
                write_port.dat_w.eq(self.write_data),
                If(we_granularity,
                    write_port.we.eq(Replicate(self.write, data_width//8) & ~self.write_mask),
                ).Else(
                    write_port.we.eq(self.write),
                ),
                If(self.read,
                    read_port.adr.eq(rdaddr),
                    self.read_data.eq(read_port.dat_r)
                )
            )
        ]

# DFI Phase Model ----------------------------------------------------------------------------------

class DFIPhaseModel(Module):
    def __init__(self, dfi, n):
        phase = getattr(dfi, "p"+str(n))

        self.bank         = phase.bank
        self.address      = phase.address

        self.wrdata       = phase.wrdata
        self.wrdata_mask  = phase.wrdata_mask

        self.rddata       = phase.rddata
        self.rddata_valid = phase.rddata_valid

        self.activate     = Signal()
        self.precharge    = Signal()
        self.write        = Signal()
        self.read         = Signal()

        # # #

        self.comb += [
            If(~phase.cs_n & ~phase.ras_n & phase.cas_n,
                self.activate.eq(phase.we_n),
                self.precharge.eq(~phase.we_n)
            ),
            If(~phase.cs_n & phase.ras_n & ~phase.cas_n,
                self.write.eq(~phase.we_n),
                self.read.eq(phase.we_n)
            )
        ]

# DFI Timings Checker ------------------------------------------------------------------------------

class SDRAMCMD:
    def __init__(self, name: str, enc: int, idx: int):
        self.name = name
        self.enc  = enc
        self.idx  = idx


class TimingRule:
    def __init__(self, prev: str, curr: str, delay: int):
        self.name  = prev + "->" + curr
        self.prev  = prev
        self.curr  = curr
        self.delay = delay


class DFITimingsChecker(Module):
    CMDS = [
        # Name, cs & ras & cas & we value
        ("PRE",  "0010"), # Precharge
        ("REF",  "0001"), # Self refresh
        ("ACT",  "0011"), # Activate
        ("RD",   "0101"), # Read
        ("WR",   "0100"), # Write
        ("ZQCS", "0110"), # ZQCS
    ]

    RULES = [
        # tRP
        ("PRE",  "ACT", "tRP"),
        ("PRE",  "REF", "tRP"),
        # tRCD
        ("ACT",  "WR",  "tRCD"),
        ("ACT",  "RD",  "tRCD"),
        # tRAS
        ("ACT",  "PRE", "tRAS"),
        # tRFC
        ("REF",  "PRE", "tRFC"),
        ("REF",  "ACT", "tRFC"),
        # tCCD
        ("WR",   "RD",  "tCCD"),
        ("WR",   "WR",  "tCCD"),
        ("RD",   "RD",  "tCCD"),
        ("RD",   "WR",  "tCCD"),
        # tRC
        ("ACT",  "ACT", "tRC"),
        # tWR
        ("WR",   "PRE", "tWR"),
        # tWTR
        ("WR",   "RD",  "tWTR"),
        # tZQCS
        ("ZQCS", "ACT", "tZQCS"),
    ]

    def add_cmds(self):
        self.cmds = {}
        for idx, (name, pattern) in enumerate(self.CMDS):
            self.cmds[name] = SDRAMCMD(name, int(pattern, 2), idx)

    def add_rule(self, prev, curr, delay):
        if not isinstance(delay, int):
            delay = self.timings[delay]
        self.rules.append(TimingRule(prev, curr, delay))

    def add_rules(self):
        self.rules = []
        for rule in self.RULES:
            self.add_rule(*rule)

    # Convert ns to ps
    def ns_to_ps(self, val):
        return int(val * 1e3)

    def ck_ns_to_ps(self, val, tck):
        c, t = val
        c = 0 if c is None else c * tck
        t = 0 if t is None else t
        return self.ns_to_ps(max(c, t))

    def prepare_timings(self, timings, refresh_mode, memtype):
        self.timings = timings
        new_timings  = {}

        tck = self.timings["tCK"]

        for key, val in self.timings.items():
            if val is None:
                val = 0
            elif key == "tCK":
                val = self.ns_to_ps(val)
            else:
                val = self.ck_ns_to_ps(val, tck)

            new_timings[key] = val

        new_timings["tRC"] = new_timings["tRAS"] + new_timings["tRP"]

        # Adjust timings relative to write burst - tWR & tWTR
        wrburst = burst_lengths[memtype] if memtype == "SDR" else burst_lengths[memtype] // 2
        wrburst = (new_timings["tCK"] * (wrburst - 1))
        new_timings["tWR"]  = new_timings["tWR"]  + wrburst
        new_timings["tWTR"] = new_timings["tWTR"] + wrburst

        self.timings = new_timings

    def __init__(self, dfi, nbanks, nphases, timings, refresh_mode, memtype, verbose=False):
        self.logging_enabled = Signal(reset=1)

        self.prepare_timings(timings, refresh_mode, memtype)
        self.add_cmds()
        self.add_rules()

        cnt = Signal(64)
        self.sync += cnt.eq(cnt + nphases)

        phases = [getattr(dfi, "p" + str(n)) for n in range(nphases)]

        last_cmd_ps = [[Signal.like(cnt) for _ in range(len(self.cmds))] for _ in range(nbanks)]
        last_cmd    = [Signal(4) for i in range(nbanks)]

        act_ps   = Array([Signal().like(cnt) for i in range(4)])
        act_curr = Signal(max=4)

        ref_issued = Signal(nphases)

        for np, phase in enumerate(phases):
            ps = Signal().like(cnt)
            self.comb += ps.eq((cnt + np)*self.timings["tCK"])
            state = Signal(4)
            self.comb += state.eq(Cat(phase.we_n, phase.cas_n, phase.ras_n, phase.cs_n))
            all_banks = Signal()

            self.comb += all_banks.eq(
                (self.cmds["REF"].enc == state) |
                ((self.cmds["PRE"].enc == state) & phase.address[10])
            )

            # tREFI
            self.comb += ref_issued[np].eq(self.cmds["REF"].enc == state)

            # Print debug information
            if verbose:
                for _, cmd in self.cmds.items():
                    self.sync += [
                        If((state == cmd.enc) & self.logging_enabled,
                            If(all_banks,
                                Display("[%016dps] P%0d " + cmd.name, ps, np)
                            ).Else(
                                Display("[%016dps] P%0d B%0d " + cmd.name, ps, np, phase.bank)
                            )
                        )
                    ]

            # Bank command monitoring
            for i in range(nbanks):
                for _, curr in self.cmds.items():
                    cmd_recv = Signal()
                    self.comb += cmd_recv.eq(((phase.bank == i) | all_banks) & (state == curr.enc))

                    # Checking rules from self.rules
                    for _, prev in self.cmds.items():
                        for rule in self.rules:
                            if rule.prev == prev.name and rule.curr == curr.name:
                                self.sync += [
                                    If(self.logging_enabled & cmd_recv & (last_cmd[i] == prev.enc) &
                                       (ps < (last_cmd_ps[i][prev.idx] + rule.delay)),
                                        Display("[%016dps] {} violation on bank %0d".format(rule.name), ps, i)
                                    )
                                ]

                    # Save command timestamp in an array
                    self.sync += If(cmd_recv, last_cmd_ps[i][curr.idx].eq(ps), last_cmd[i].eq(state))

                    # tRRD & tFAW
                    if curr.name == "ACT":
                        act_next = Signal().like(act_curr)
                        self.comb += act_next.eq(act_curr+1)

                        # act_curr points to newest ACT timestamp
                        self.sync += [
                            If(self.logging_enabled & cmd_recv & (ps < (act_ps[act_curr] + self.timings["tRRD"])),
                                Display("[%016dps] tRRD violation on bank %0d", ps, i)
                            )
                        ]

                        # act_next points to the oldest ACT timestamp
                        self.sync += [
                            If(self.logging_enabled & cmd_recv & (ps < (act_ps[act_next] + self.timings["tFAW"])),
                                Display("[%016dps] tFAW violation on bank %0d", ps, i)
                            )
                        ]

                        # Save ACT timestamp in a circular buffer
                        self.sync += If(cmd_recv, act_ps[act_next].eq(ps), act_curr.eq(act_next))

        # tREFI
        ref_ps      = Signal().like(cnt)
        ref_ps_mod  = Signal().like(cnt)
        ref_ps_diff = Signal(min=-2**63, max=2**63)
        curr_diff   = Signal().like(ref_ps_diff)

        self.comb += curr_diff.eq(ps - (ref_ps + self.timings["tREFI"]))

        # Work in 64ms periods
        self.sync += [
            If(ref_ps_mod < int(64e9),
                ref_ps_mod.eq(ref_ps_mod + nphases * self.timings["tCK"])
            ).Else(
                ref_ps_mod.eq(0)
            )
        ]

        # Update timestamp and difference
        self.sync += If(ref_issued != 0, ref_ps.eq(ps), ref_ps_diff.eq(ref_ps_diff - curr_diff))

        self.sync += [
            If((self.logging_enabled & ref_ps_mod == 0) & (ref_ps_diff > 0),
                Display("[%016dps] tREFI violation (64ms period): %0d", ps, ref_ps_diff)
            )
        ]

        # Report any refresh periods longer than tREFI
        if verbose:
            ref_done = Signal()
            self.sync += [
                If(ref_issued != 0,
                    ref_done.eq(1),
                    If(self.logging_enabled & ~ref_done,
                        Display("[%016dps] Late refresh", ps)
                    )
                )
            ]

            self.sync += [
                If(self.logging_enabled & (curr_diff > 0) & ref_done & (ref_issued == 0),
                    Display("[%016dps] tREFI violation", ps),
                    ref_done.eq(0)
                )
            ]

        # There is a maximum delay between refreshes on >=DDR
        ref_limit = {"1x": 9, "2x": 17, "4x": 36}
        if memtype != "SDR":
            refresh_mode = "1x" if refresh_mode is None else refresh_mode
            ref_done = Signal()
            self.sync += If(ref_issued != 0, ref_done.eq(1))
            self.sync += [
                If(self.logging_enabled & (ref_issued == 0) & ref_done &
                   (ref_ps > (ps + ref_limit[refresh_mode] * self.timings['tREFI'])),
                    Display("[%016dps] tREFI violation (too many postponed refreshes)", ps),
                    ref_done.eq(0)
                )
            ]

# SDRAM PHY Settings -------------------------------------------------------------------------------

sdram_module_nphases = {
    "SDR":   1,
    "DDR":   2,
    "LPDDR": 2,
    "DDR2":  2,
    "DDR3":  4,
    "DDR4":  4,
}

def get_sdram_phy_settings(memtype, data_width, clk_freq):
    nphases = sdram_module_nphases[memtype]

    if memtype == "SDR":
        # Settings from gensdrphy
        rdphase       = 0
        wrphase       = 0
        cl            = 2
        cwl           = None
        read_latency  = 4
        write_latency = 0
    elif memtype in ["DDR", "LPDDR"]:
        # Settings from s6ddrphy
        rdphase       = 0
        wrphase       = 1
        cl            = 3
        cwl           = None
        read_latency  = 5
        write_latency = 0
    elif memtype in ["DDR2", "DDR3"]:
        # Settings from s7ddrphy
        tck             = 2/(2*nphases*clk_freq)
        cl, cwl         = get_default_cl_cwl(memtype, tck)
        cl_sys_latency  = get_sys_latency(nphases, cl)
        cwl_sys_latency = get_sys_latency(nphases, cwl)
        rdphase         = get_sys_phase(nphases, cl_sys_latency, cl)
        wrphase         = get_sys_phase(nphases, cwl_sys_latency, cwl)
        read_latency    = cl_sys_latency + 6
        write_latency   = cwl_sys_latency - 1
    elif memtype == "DDR4":
        # Settings from usddrphy
        tck             = 2/(2*nphases*clk_freq)
        cl, cwl         = get_default_cl_cwl(memtype, tck)
        cl_sys_latency  = get_sys_latency(nphases, cl)
        cwl_sys_latency = get_sys_latency(nphases, cwl)
        rdphase         = get_sys_phase(nphases, cl_sys_latency, cl)
        wrphase         = get_sys_phase(nphases, cwl_sys_latency, cwl)
        read_latency    = cl_sys_latency + 5
        write_latency   = cwl_sys_latency - 1

    sdram_phy_settings = {
        "nphases":       nphases,
        "rdphase":       rdphase,
        "wrphase":       wrphase,
        "cl":            cl,
        "cwl":           cwl,
        "read_latency":  read_latency,
        "write_latency": write_latency,
    }

    return PhySettings(
        phytype      = "SDRAMPHYModel",
        memtype      = memtype,
        databits     = data_width,
        dfi_databits = data_width if memtype == "SDR" else 2*data_width,
        **sdram_phy_settings,
    )

# SDRAM PHY Model ----------------------------------------------------------------------------------

class SDRAMPHYModel(Module):
    def __prepare_bank_init_data(self, init, nbanks, nrows, ncols, data_width, address_mapping):
        mem_size          = (self.settings.databits//8)*(nrows*ncols*nbanks)
        bank_size         = mem_size // nbanks
        column_size       = bank_size // nrows
        model_bank_size   = bank_size // (data_width//8)
        model_column_size = model_bank_size // nrows
        model_data_ratio  = data_width // 32
        data_width_bytes  = data_width // 8
        bank_init         = [[] for i in range(nbanks)]

        # Pad init if too short
        if len(init)%data_width_bytes != 0:
            init.extend([0]*(data_width_bytes-len(init)%data_width_bytes))


        # Convert init data width from 32-bit to data_width if needed
        if model_data_ratio > 1:
            new_init = [0]*(len(init)//model_data_ratio)
            for i in range(0, len(init), model_data_ratio):
                ints = init[i:i+model_data_ratio]
                strs = "".join("{:08x}".format(x) for x in reversed(ints))
                new_init[i//model_data_ratio] = int(strs, 16)
            init = new_init
        elif model_data_ratio == 0:
            assert data_width_bytes in [1, 2]
            model_data_ratio = 4 // data_width_bytes
            struct_unpack_patterns = {1: "4B", 2: "2H"}
            new_init = [0]*int(len(init)*model_data_ratio)
            for i in range(len(init)):
                new_init[model_data_ratio*i:model_data_ratio*(i+1)] = struct.unpack(
                    struct_unpack_patterns[data_width_bytes],
                    struct.pack("I", init[i])
                )[0:model_data_ratio]
            init = new_init

        if address_mapping == "ROW_BANK_COL":
            for row in range(nrows):
                for bank in range(nbanks):
                    start = (row*nbanks*model_column_size + bank*model_column_size)
                    end   = min(start + model_column_size, len(init))
                    if start > len(init):
                        break
                    bank_init[bank].extend(init[start:end])
        elif address_mapping == "BANK_ROW_COL":
            for bank in range(nbanks):
                start = bank*model_bank_size
                end   = min(start + model_bank_size, len(init))
                if start > len(init):
                    break
                bank_init[bank] = init[start:end]

        return bank_init

    def __init__(self, module, settings=None, data_width=None, clk_freq=100e6,
        we_granularity         = 8,
        init                   = [],
        address_mapping        = "ROW_BANK_COL",
        verbosity              = SDRAM_VERBOSE_OFF):

        # PHY Settings -----------------------------------------------------------------------------
        if settings is None:
            assert data_width is not None
            settings = get_sdram_phy_settings(
                memtype    = module.memtype,
                data_width = data_width,
                clk_freq   = clk_freq
            )

        # Parameters -------------------------------------------------------------------------------
        burst_length = {
            "SDR":   1,
            "DDR":   2,
            "LPDDR": 2,
            "DDR2":  2,
            "DDR3":  2,
            "RPC":   4,
            "DDR4":  2,
            }[settings.memtype]

        addressbits   = module.geom_settings.addressbits
        bankbits      = module.geom_settings.bankbits
        rowbits       = module.geom_settings.rowbits
        colbits       = module.geom_settings.colbits

        self.settings = settings
        self.module   = module

        # DFI Interface ----------------------------------------------------------------------------
        self.dfi = Interface(
            addressbits = addressbits,
            bankbits    = bankbits,
            nranks      = self.settings.nranks,
            databits    = self.settings.dfi_databits,
            nphases     = self.settings.nphases
        )

        # # #

        nphases    = self.settings.nphases
        nbanks     = 2**bankbits
        nrows      = 2**rowbits
        ncols      = 2**colbits
        data_width = self.settings.dfi_databits*self.settings.nphases

        # DFI phases -------------------------------------------------------------------------------
        phases = [DFIPhaseModel(self.dfi, n) for n in range(self.settings.nphases)]
        self.submodules += phases

        # DFI timing checker -----------------------------------------------------------------------
        if verbosity > SDRAM_VERBOSE_OFF:
            timings = {"tCK": (1e9 / clk_freq) / nphases}
            CK_NS = ["tRFC", "tWTR", "tFAW", "tCCD", "tRRD", "tZQCS"]
            REF = ["tREFI", "tRFC"]

            for name in _speedgrade_timings + _technology_timings:
                key = self.module.timing_settings.fine_refresh_mode if name in REF else None
                timings[name] = self.module.get(name, key)

            timing_checker = DFITimingsChecker(
                dfi          = self.dfi,
                nbanks       = nbanks,
                nphases      = nphases,
                timings      = timings,
                refresh_mode = self.module.timing_settings.fine_refresh_mode,
                memtype      = settings.memtype,
                verbose      = verbosity > SDRAM_VERBOSE_DBG)
            self.submodules += timing_checker

        # Bank init data ---------------------------------------------------------------------------
        bank_init  = [None for i in range(nbanks)]

        if init:
            bank_init = self.__prepare_bank_init_data(
                init            = init,
                nbanks          = nbanks,
                nrows           = nrows,
                ncols           = ncols,
                data_width      = data_width,
                address_mapping = address_mapping
            )

        # Banks ------------------------------------------------------------------------------------
        banks = [BankModel(
            data_width     = data_width,
            nrows          = nrows,
            ncols          = ncols,
            burst_length   = burst_length,
            nphases        = nphases,
            we_granularity = we_granularity,
            init           = bank_init[i]) for i in range(nbanks)]
        self.submodules += banks

        # Connect DFI phases to Banks (CMDs, Write datapath) ---------------------------------------
        for nb, bank in enumerate(banks):
            # Bank activate
            activates = Signal(len(phases))
            cases     = {}
            for np, phase in enumerate(phases):
                self.comb += activates[np].eq(phase.activate)
                cases[2**np] = [
                    bank.activate.eq(phase.bank == nb),
                    bank.activate_row.eq(phase.address)
                ]
            self.comb += Case(activates, cases)

            # Bank precharge
            precharges = Signal(len(phases))
            cases      = {}
            for np, phase in enumerate(phases):
                self.comb += precharges[np].eq(phase.precharge)
                cases[2**np] = [
                    bank.precharge.eq((phase.bank == nb) | phase.address[10])
                ]
            self.comb += Case(precharges, cases)

            # Bank writes
            bank_write = Signal()
            bank_write_col = Signal(max=ncols)
            writes = Signal(len(phases))
            cases  = {}
            for np, phase in enumerate(phases):
                self.comb += writes[np].eq(phase.write)
                cases[2**np] = [
                    bank_write.eq(phase.bank == nb),
                    bank_write_col.eq(phase.address)
                ]
            self.comb += Case(writes, cases)
            self.comb += [
                bank.write_data.eq(Cat(*[phase.wrdata for phase in phases])),
                bank.write_mask.eq(Cat(*[phase.wrdata_mask for phase in phases]))
            ]

            # Simulate write latency
            for i in range(self.settings.write_latency):
                new_bank_write     = Signal()
                new_bank_write_col = Signal(max=ncols)
                self.sync += [
                    new_bank_write.eq(bank_write),
                    new_bank_write_col.eq(bank_write_col)
                ]
                bank_write = new_bank_write
                bank_write_col = new_bank_write_col

            self.comb += [
                bank.write.eq(bank_write),
                bank.write_col.eq(bank_write_col)
            ]

            # Bank reads
            reads = Signal(len(phases))
            cases = {}
            for np, phase in enumerate(phases):
                self.comb += reads[np].eq(phase.read)
                cases[2**np] = [
                    bank.read.eq(phase.bank == nb),
                    bank.read_col.eq(phase.address)
            ]
            self.comb += Case(reads, cases)

        # Connect Banks to DFI phases (CMDs, Read datapath) ----------------------------------------
        banks_read      = Signal()
        banks_read_data = Signal(data_width)
        self.comb += [
            banks_read.eq(reduce(or_, [bank.read for bank in banks])),
            banks_read_data.eq(reduce(or_, [bank.read_data for bank in banks]))
        ]

        # Simulate read latency --------------------------------------------------------------------
        for i in range(self.settings.read_latency):
            new_banks_read      = Signal()
            new_banks_read_data = Signal(data_width)
            self.sync += [
                new_banks_read.eq(banks_read),
                new_banks_read_data.eq(banks_read_data)
            ]
            banks_read      = new_banks_read
            banks_read_data = new_banks_read_data

        self.comb += [
            Cat(*[phase.rddata_valid for phase in phases]).eq(banks_read),
            Cat(*[phase.rddata for phase in phases]).eq(banks_read_data)
        ]



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

class S7Common(Module):
    def idelaye2(self, *, din, dout, init=0, rst=None, inc=None, clk="sys2x"):
        assert not ((rst is None) ^ (inc is None))
        fixed = rst is None

        params = dict(
            p_SIGNAL_PATTERN        = "DATA",
            p_DELAY_SRC             = "IDATAIN",
            p_CINVCTRL_SEL          = "FALSE",
            p_HIGH_PERFORMANCE_MODE = "TRUE",
            p_REFCLK_FREQUENCY      = self.iodelay_clk_freq/1e6,
            p_PIPE_SEL              = "FALSE",
            p_IDELAY_VALUE          = init,
            p_IDELAY_TYPE           = "FIXED",
            i_IDATAIN  = din,
            o_DATAOUT  = dout,
        )

        if not fixed:
            params.update(dict(
                p_IDELAY_TYPE  = "VARIABLE",
                i_C        = ClockSignal(clk),  # must be same as in ODELAYE2
                i_LD       = rst,
                i_CE       = inc,
                i_LDPIPEEN = 0,
                i_INC      = 1,
            ))

        self.specials += Instance("IDELAYE2", **params)

    def odelaye2(self, *, din, dout, init=0, rst=None, inc=None, clk="sys2x"):  # Not available for Artix7
        assert not ((rst is None) ^ (inc is None))
        fixed = rst is None

        params = dict(
            p_SIGNAL_PATTERN        = "DATA",
            p_DELAY_SRC             = "ODATAIN",
            p_CINVCTRL_SEL          = "FALSE",
            p_HIGH_PERFORMANCE_MODE = "TRUE",
            p_REFCLK_FREQUENCY      = self.iodelay_clk_freq/1e6,
            p_PIPE_SEL              = "FALSE",
            p_ODELAY_VALUE          = init,
            p_ODELAY_TYPE           = "FIXED",
            i_ODATAIN  = din,
            o_DATAOUT  = dout,
        )

        if not fixed:
            params.update(dict(
                p_ODELAY_TYPE  = "VARIABLE",
                i_C        = ClockSignal(clk),  # must be same as CLKDIV in OSERDESE2
                i_LD       = rst,
                i_CE       = inc,
                i_LDPIPEEN = 0,
                i_INC      = 1,
            ))

        self.specials += Instance("ODELAYE2", **params)

    def oserdese2_ddr(self, *, din, clk, dout=None, dout_fb=None, tin=None, tout=None, clkdiv="sys2x"):
        data_width = len(din)
        assert data_width == 8, (data_width, din)
        assert not ((tin is None) ^ (tout is None)), "When using tristate specify both `tin` and `tout`"
        assert not ((dout is None) and (dout_fb is None)), "Output to OQ (-> IOB) and/or to OFB (-> ISERDESE2/ODELAYE2)"

        dout = Signal() if dout is None else dout
        dout_fb = Signal() if dout_fb is None else dout_fb

        params = dict(
            p_SERDES_MODE    = "MASTER",
            p_DATA_WIDTH     = data_width,
            p_TRISTATE_WIDTH = 1,
            p_DATA_RATE_OQ   = "DDR",
            p_DATA_RATE_TQ   = "BUF",
            i_RST    = ResetSignal() | self._rst.storage,
            i_CLK    = ClockSignal(clk),
            i_CLKDIV = ClockSignal(clkdiv),
            o_OQ     = dout,
            o_OFB    = dout_fb,
            i_OCE    = 1,
        )

        for i in range(data_width):
            params[f"i_D{i+1}"] = din[i]

        if tin is not None:
            # with DATA_RATE_TQ=BUF tristate is asynchronous, so it should be delayed by OSERDESE2 latency
            params.update(dict(i_TCE=1, i_T1=tin, o_TQ=tout))

        self.specials += Instance("OSERDESE2", **params)

    def oserdese2_sdr(self, **kwargs):
        # Use 8:1 OSERDESE2 DDR instead of 4:1 OSERDESE2 SDR to have the same latency
        din = kwargs["din"]
        data_width = len(din)
        assert data_width in [1, 2, 4]
        ratio = 8 // data_width
        din_ddr = Signal(8)
        kwargs["din"] = din_ddr
        self.comb += din_ddr.eq(Cat(*[Replicate(bit, ratio) for bit in din]))
        self.oserdese2_ddr(**kwargs)

    def iserdese2_ddr(self, *, din, dout, clk, clkdiv="sys2x"):
        data_width = len(dout)
        assert data_width == 8, (data_width, dout)

        params = dict(
            p_SERDES_MODE    = "MASTER",
            p_INTERFACE_TYPE = "NETWORKING",
            p_DATA_WIDTH     = data_width,
            p_DATA_RATE      = "DDR",
            p_NUM_CE         = 1,
            p_IOBDELAY       = "IFD",
            i_RST     = ResetSignal() | self._rst.storage,
            i_CLK     = ClockSignal(clk),
            i_CLKB    = ~ClockSignal(clk),
            i_CLKDIV  = ClockSignal(clkdiv),
            i_BITSLIP = 0,
            i_CE1     = 1,
            i_DDLY    = din,
        )

        for i in range(data_width):
            # invert order
            params[f"o_Q{i+1}"] = dout[(data_width - 1) - i]

        self.specials += Instance("ISERDESE2", **params)

    def iserdese2_sdr(self, **kwargs):
        dout = kwargs["dout"]
        data_width = len(dout)
        assert data_width in [1, 2, 4]
        ratio = 8 // data_width
        dout_ddr = Signal(8)
        kwargs["dout"] = dout_ddr
        self.comb += dout.eq(Cat(*[dout_ddr[bit] for bit in range(0, 8, ratio)]))
        self.iserdese2_ddr(**kwargs)

    def obufds(self, *, din, dout, dout_b):
        self.specials += Instance("OBUFDS",
            i_I  = din,
            o_O  = dout,
            o_OB = dout_b,
        )

    def iobufds(self, *, din, dout, dinout, dinout_b, tin):
        self.specials += Instance("IOBUFDS",
            i_T    = tin,
            i_I    = din,
            o_O    = dout,
            io_IO  = dinout,
            io_IOB = dinout_b,
        )

    def iobuf(self, *, din, dout, dinout, tin):
        self.specials += Instance("IOBUF",
            i_T   = tin,
            i_I   = din,
            o_O   = dout,
            io_IO = dinout,
        )



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2020-2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.fhdl.specials import Tristate

from litex.soc.interconnect.csr import CSR

from litedram.phy.rpc.basephy import BasePHY


class SimulationPHY(BasePHY):
    def __init__(self, *args, generate_read_data=True, **kwargs):
        kwargs.update(dict(
            write_ser_latency = 1,
            read_des_latency  = 0,
            phytype           = "RPC" + self.__class__.__name__,
        ))
        super().__init__(*args, **kwargs)

        self.generate_read_data = generate_read_data

        # fake delays (make no nsense in simulation, but sdram.c expects them)
        self.settings.read_leveling = True
        self.settings.delays = 1
        self._rdly_dq_rst = CSR()
        self._rdly_dq_inc = CSR()

        # For simulation purpose (serializers/deserializers)
        self.sd_ddr_90  = getattr(self.sync, "sys4x_90_ddr")
        self.sd_ddr_180 = getattr(self.sync, "sys4x_180_ddr")

    def do_clock_serialization(self, clk_1ck_out, clk_p, clk_n):
        ser = Serializer(self.sync, self.sd_ddr_180, clk_1ck_out, delay=1)
        self.submodules += ser
        self.comb += clk_p.eq(ser.o)
        self.comb += clk_n.eq(~ser.o)

    def do_stb_serialization(self, stb_1ck_out, stb):
        ser = Serializer(self.sync, self.sd_ddr_90, stb_1ck_out, delay=1)
        self.submodules += ser
        self.comb += stb.eq(ser.o)

    def do_dqs_serialization(self, dqs_1ck_out, dqs_1ck_in, dqs_oe, dqs_p, dqs_n):
        # Delay dqs_oe by 1 cycle (Serializer latency) and register it on output domain
        dqs_oe_d   = Signal()
        dqs_oe_cdc = Signal()
        self.sync += dqs_oe_d.eq(dqs_oe)
        self.sync.sys4x_180 += dqs_oe_cdc.eq(dqs_oe_d)

        for i in range(len(dqs_p)):
            dqs_out = Signal()
            dqs_in  = Signal()  # TODO: use it for reading

            ser = Serializer(self.sync, self.sd_ddr_180, dqs_1ck_out, delay=1)
            self.submodules += ser
            self.comb += dqs_out.eq(ser.o)

            if i == 0:
                des = Deserializer(self.sd_ddr_180, dqs_in, dqs_1ck_in)
                self.submodules += des

            # self.specials += Tristate(dqs_p[i],  dqs_out, dqs_oe_cdc, dqs_in)
            # self.specials += Tristate(dqs_n[i], ~dqs_out, dqs_oe_cdc)
            self.comb += [
                If(dqs_oe_cdc,
                    dqs_p[i].eq(dqs_out),
                    dqs_n[i].eq(~dqs_out),
                ),
                dqs_in.eq(dqs_p[i]),
            ]

    def do_db_serialization(self, db_1ck_out, db_1ck_in, db_oe, db):
        if self.generate_read_data:
            # Dummy read data generator for simulation purpose
            dq_in_dummy = Signal(self.databits)
            gen = DummyReadGenerator(dq_in=db, dq_out=dq_in_dummy, stb_in=self.pads.stb,
                                     cl=self.settings.cl)
            self.submodules += ClockDomainsRenamer({"sys": "sys4x_180_ddr"})(gen)

        # Delay db_oe by 1 cycle (Serializer latency) and register it on output domain
        db_oe_d   = Signal()
        db_oe_cdc = Signal()
        self.sync += db_oe_d.eq(db_oe)
        self.sync.sys4x_90 += db_oe_cdc.eq(db_oe_d)

        for i in range(self.databits):
            # To/from tristate
            dq_out = Signal()
            dq_in = Signal()

            ser = Serializer(self.sync, self.sd_ddr_90, db_1ck_out[i], delay=1)
            self.submodules += ser
            self.comb += dq_out.eq(ser.o)

            # Use sd_ddr_90 in simulation, in general leveling would be needed.
            if self.generate_read_data:
                des = Deserializer(self.sd_ddr_90, dq_in_dummy[i], db_1ck_in[i])
            else:
                des = Deserializer(self.sd_ddr_90, dq_in, db_1ck_in[i])
            self.submodules += des

            # self.specials += Tristate(db[i], dq_out, db_oe_cdc, dq_in)
            self.comb += [
                If(db_oe_cdc, db[i].eq(dq_out)),
                dq_in.eq(db[i]),
            ]

    def do_cs_serialization(self, cs_n_1ck_out, cs_n):
        ser = Serializer(self.sync, self.sd_ddr_90, cs_n_1ck_out, delay=1)
        self.submodules += ser
        self.comb += cs_n.eq(ser.o)

# Simulation of READ ouput -------------------------------------------------------------------------

class DummyReadGenerator(Module):
    def __init__(self, stb_in, dq_in, dq_out, cl):
        # self.sync should be a 4x DDR clock phase-aligned with CLK

        data_counter = Signal(max=16)
        cmd_counter = Signal(max=16)
        # count STB zeros, 2 full-rate cycles (4 DDR) mean STB preamble, but more mean RESET
        stb_zero_counter = Signal(max=4 + 2)
        self.sync += \
            If(stb_in == 0,
                If(stb_zero_counter != 2**len(stb_zero_counter) - 1,
                    stb_zero_counter.eq(stb_zero_counter + 1)
                )
            ).Else(
                stb_zero_counter.eq(0)
            )

        # Because the generator clock domain is phase shifted in relation to DB, we have to
        # register the value to hold it until the end of our clock, so that we get correct FSM
        # transitions
        # FIXME: would cause problems if clocks were aligned
        dq_in_r = Signal.like(dq_in)
        self.sync += dq_in_r.eq(dq_in)

        # generate DQS just for viewing signal dump
        dqs_out = Signal()
        dqs_oe  = Signal()

        data = Array([
            0x0000,  # 0x0110,
            0x1111,  # 0x1221,
            0x2222,  # 0x2332,
            0x3333,  # 0x3443,
            0x4444,  # 0x4554,
            0x5555,  # 0x5665,
            0x6666,  # 0x6776,
            0x7777,  # 0x7887,
            0x8888,  # 0x8998,
            0x9999,  # 0x9aa9,
            0xaaaa,  # 0xabba,
            0xbbbb,  # 0xbccb,
            0xcccc,  # 0xcddc,
            0xdddd,  # 0xdeed,
            0xeeee,  # 0xeffe,
            0xffff,  # 0xf00f,
        ])

        self.submodules.fsm = fsm = FSM()
        fsm.act("IDLE",
            NextValue(data_counter, 0),
            If(stb_zero_counter == 4,
                NextState("CHECK_CMD_P")
            )
        )
        fsm.act("CHECK_CMD_P",
            If(dq_in_r[:3] == 0,
                NextState("CHECK_CMD_N")
            ).Else(
                NextState("IDLE")
            )
        )
        fsm.act("CHECK_CMD_N",
            If(dq_in_r[0] == 0,
                NextState("CL_WAIT")
            ).Else(
                NextState("IDLE")
            )
        )
        # 2x for DDR, -1 for CHECK_CMD_*, -2 for preamble
        fsm.delayed_enter("CL_WAIT", "PREAMBLE_P", 2*(cl - 1) - 2)
        fsm.act("PREAMBLE_P",
            dqs_oe.eq(1),
            dqs_out.eq(0),
            NextState("PREAMBLE_N")
        )
        fsm.act("PREAMBLE_N",
            dqs_oe.eq(1),
            dqs_out.eq(0),
            NextState("SEND_DATA")
        )
        fsm.act("SEND_DATA",
            dqs_oe.eq(1),
            dqs_out.eq(data_counter[0] == 0),
            dq_out.eq(data[data_counter + cmd_counter]),
            NextValue(data_counter, data_counter + 1),
            If(data_counter == 16 - 1,
                NextValue(cmd_counter, cmd_counter + 1),
                NextState("IDLE")
            )
        )

# I/O Primitives -----------------------------------------------------------------------------------

class Serializer(Module):
    """Serialize input signals into one output with 1 sd_clk clock latency"""
    def __init__(self, sd_clk, sd_clkdiv, inputs, reset=0, name=None, delay=0):
        # `delay` must be used to get correct data_cntr=0 for phase delayed signals
        assert(len(inputs) > 0)
        assert(len(s) == len(inputs[0]) for s in inputs)

        data_width = len(inputs)
        signal_width = len(inputs[0])
        cntr_max = 2 * data_width

        if not isinstance(inputs, Array):
            inputs = Array(inputs)

        self.o = Signal(signal_width)
        data_cntr = Signal(max=cntr_max, name=name, reset=(0 - delay) % cntr_max)
        sd_clkdiv += If(reset, data_cntr.eq(0)).Else(data_cntr.eq(data_cntr+1))

        # If we used inputs combinatorically, then we will have a problem when the serialization
        # clock is not phase-aligned with the `inputs` clock, e.g.
        #  inputs clk 1:      | 0 : 1 : 2 : 3 | 4 : 5 : 6 : 7 |
        #  serialized 90 deg:   | 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 |
        # Here input 3 becomes invalid before serialized output moves to 4, so it would actually
        # send 3 for the 1st half and 7 for the 2nd half. To avoid this we introduce 1 clock
        # latency of the inputs and use 2 registers to hold the values of subsequent inputs.
        inputs_cnt = Signal()
        inputs_d   = Array([Signal.like(i) for i in [*inputs, *inputs]])
        sd_clk += [
            inputs_cnt.eq(inputs_cnt + 1),
            Case(inputs_cnt, {
                0: [i_d.eq(i) for i_d, i in zip(inputs_d[data_width:], inputs)],
                1: [i_d.eq(i) for i_d, i in zip(inputs_d[:data_width], inputs)],
            })
        ]

        self.comb += self.o.eq(inputs_d[data_cntr])


class Deserializer(Module):
    """Deserialize an input signal into outputs in the `sd` clock domain"""
    def __init__(self, sd, input, outputs, reset=0, name=None):
        assert(len(outputs) > 0)
        assert(len(s) == len(outputs[0]) for s in outputs)
        assert(len(outputs[0]) == len(input))

        data_width = len(outputs)
        signal_width = len(outputs[0])

        if not isinstance(outputs, Array):
            outputs = Array(outputs)

        data_cntr = Signal(log2_int(data_width), name=name)
        sd += If(reset, data_cntr.eq(0)).Else(data_cntr.eq(data_cntr+1))
        sd += outputs[data_cntr].eq(input)



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2015 Sebastien Bourdeauducq <sb@m-labs.hk>
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2018 John Sully <john@csquare.ca>
# Copyright (c) 2019 Ambroz Bizjak <abizjak.pro@gmail.com>
# Copyright (c) 2019 Antony Pavlov <antonynpavlov@gmail.com>
# Copyright (c) 2018 bunnie <bunnie@kosagi.com>
# Copyright (c) 2018 David Shah <dave@ds0.me>
# Copyright (c) 2019 Steve Haynal - VSD Engineering
# Copyright (c) 2018 Tim 'mithro' Ansell <me@mith.ro>
# Copyright (c) 2018 Daniel Kucera <daniel.kucera@gmail.com>
# Copyright (c) 2018 Mikołaj Sowiński <mikolaj.sowinski@gmail.com>
# Copyright (c) 2020-2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from math import ceil
from collections import namedtuple

from migen import *

from litedram.common import Settings, GeomSettings, TimingSettings

# Timings ------------------------------------------------------------------------------------------

_technology_timings = ["tREFI", "tWTR", "tCCD", "tRRD", "tZQCS"]

class _TechnologyTimings(Settings):
    def __init__(self, tREFI, tWTR, tCCD, tRRD, tZQCS=None):
        self.set_attributes(locals())


_speedgrade_timings = ["tRP", "tRCD", "tWR", "tRFC", "tFAW", "tRAS"]

class _SpeedgradeTimings(Settings):
    def __init__(self, tRP, tRCD, tWR, tRFC, tFAW, tRAS):
        self.set_attributes(locals())

# SPD ----------------------------------------------------------------------------------------------

def _read_field(byte, nbits, shift):
    mask = 2**nbits - 1
    return (byte & (mask << shift)) >> shift

def _twos_complement(value, nbits):
    if value & (1 << (nbits - 1)):
        value -= (1 << nbits)
    return value

def _word(msb, lsb):
    return (msb << 8) | lsb

# most signifficant (upper) / least signifficant (lower) nibble
def _msn(byte):
    return _read_field(byte, nbits=4, shift=4)

def _lsn(byte):
    return _read_field(byte, nbits=4, shift=0)


class DDR3SPDData:
    memtype = "DDR3"
    _speedgrades = [800, 1066, 1333, 1600, 1866, 2133]

    def __init__(self, spd_data):
        self.get_geometry(spd_data)
        self.init_timebase(spd_data)
        self.get_timings(spd_data)

    def get_geometry(self, data):
        bankbits = {
            0b000: 3,
            0b001: 4,
            0b010: 5,
            0b011: 6,
        }[_read_field(data[4], nbits=3, shift=4)]
        rowbits = {
            0b000: 12,
            0b001: 13,
            0b010: 14,
            0b011: 15,
            0b100: 16,
        }[_read_field(data[5], nbits=3, shift=3)]
        colbits = {
            0b000:  9,
            0b001: 10,
            0b010: 11,
            0b011: 12,
        }[_read_field(data[5], nbits=3, shift=0)]

        self.nbanks = 2**bankbits
        self.nrows = 2**rowbits
        self.ncols = 2**colbits

    def get_timings(self, spd_data):
        b = spd_data
        tck_min  = self.txx_ns(mtb=b[12], ftb=b[34])
        taa_min  = self.txx_ns(mtb=b[16], ftb=b[35])
        twr_min  = self.txx_ns(mtb=b[17])
        trcd_min = self.txx_ns(mtb=b[18], ftb=b[36])
        trrd_min = self.txx_ns(mtb=b[19])
        trp_min  = self.txx_ns(mtb=b[20], ftb=b[37])
        tras_min = self.txx_ns(mtb=_word(_lsn(b[21]), b[22]))
        trc_min  = self.txx_ns(mtb=_word(_msn(b[21]), b[23]), ftb=b[38])
        trfc_min = self.txx_ns(mtb=_word(b[25], b[24]))
        twtr_min = self.txx_ns(mtb=b[26])
        trtp_min = self.txx_ns(mtb=b[27])
        tfaw_min = self.txx_ns(mtb=_word(_lsn(b[28]), b[29]))

        technology_timings = _TechnologyTimings(
            tREFI = 64e6/8192,      # 64ms/8192ops
            tWTR  = (4, twtr_min),  # min 4 cycles
            tCCD  = (4, None),      # min 4 cycles
            tRRD  = (4, trrd_min),  # min 4 cycles
            tZQCS = (64, 80),
        )
        speedgrade_timings = _SpeedgradeTimings(
            tRP  = trp_min,
            tRCD = trcd_min,
            tWR  = twr_min,
            tRFC = (None, trfc_min),
            tFAW = (None, tfaw_min),
            tRAS = tras_min,
        )

        self.speedgrade = str(self.speedgrade_freq(tck_min))
        self.technology_timings = technology_timings
        self.speedgrade_timings = {
            self.speedgrade: speedgrade_timings,
            "default": speedgrade_timings,
        }

    def init_timebase(self, data):
        # All the DDR3 timings are defined in the units of "timebase", which
        # consists of medium timebase (nanosec) and fine timebase (picosec).
        fine_timebase_dividend = _read_field(data[9], nbits=4, shift=4)
        fine_timebase_divisor  = _read_field(data[9], nbits=4, shift=0)
        fine_timebase_ps = fine_timebase_dividend / fine_timebase_divisor
        self.fine_timebase_ns = fine_timebase_ps * 1e-3
        medium_timebase_dividend = data[10]
        medium_timebase_divisor  = data[11]
        self.medium_timebase_ns = medium_timebase_dividend / medium_timebase_divisor

    def txx_ns(self, mtb, ftb=0):
        """Get tXX in nanoseconds from medium and (optional) fine timebase."""
        # decode FTB encoded in 8-bit two's complement
        ftb = _twos_complement(ftb, 8)
        return mtb * self.medium_timebase_ns + ftb * self.fine_timebase_ns

    @classmethod
    def speedgrade_freq(cls, tck_ns):
        # Calculate rounded speedgrade frequency from tck_min
        freq_mhz = (1 / (tck_ns * 1e-9)) / 1e6
        freq_mhz *= 2  # clock rate -> transfer rate (DDR)
        for f in cls._speedgrades:
            # Due to limited tck accuracy of 1ps, calculations may yield higher
            # frequency than in reality (e.g. for DDR3-1866: tck=1.071 ns ->
            # -> f=1867.4 MHz, while real is f=1866.6(6) MHz).
            max_error = 2
            if abs(freq_mhz - f) < max_error:
                return f
        raise ValueError("Transfer rate = {:.2f} does not correspond to any speedgrade"
                         .format(freq_mhz))


class DDR4SPDData(DDR3SPDData):
    memtype = "DDR4"
    _speedgrades = [1600, 1866, 2133, 2400, 2666, 2933, 3200]

    def get_geometry(self, data):
        bankgroupbits = {
            0b00: 0,
            0b01: 1,
            0b10: 2,
        }[_read_field(data[4], nbits=2, shift=6)]
        bankbits = {
            0b00: 2,
            0b01: 3,
        }[_read_field(data[4], nbits=2, shift=4)]
        rowbits = {
            0b000: 12,
            0b001: 13,
            0b010: 14,
            0b011: 15,
            0b100: 16,
            0b101: 17,
            0b110: 18,
        }[_read_field(data[5], nbits=3, shift=3)]
        colbits = {
            0b000:  9,
            0b001: 10,
            0b010: 11,
            0b011: 12,
        }[_read_field(data[5], nbits=3, shift=0)]

        self.ngroups = 2**bankgroupbits
        self.ngroupbanks = 2**bankbits
        self.nbanks = self.ngroups * self.ngroupbanks
        self.nrows = 2**rowbits
        self.ncols = 2**colbits

    def init_timebase(self, data):
        # there is only one possible value for DDR4
        self.medium_timebase_ns = {
            0b00: 125e-3,
        }[_read_field(data[17], nbits=2, shift=2)]
        self.fine_timebase_ns = {
            0b00: 1e-3,
        }[_read_field(data[17], nbits=2, shift=0)]

    def get_timings(self, data):
        b = data

        self.trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}

        tckavg_min  = self.txx_ns(mtb=b[18], ftb=b[125])
        tckavg_max  = self.txx_ns(mtb=b[19], ftb=b[124])
        taa_min  = self.txx_ns(mtb=b[24], ftb=b[123])
        trcd_min = self.txx_ns(mtb=b[25], ftb=b[122])
        trp_min  = self.txx_ns(mtb=b[26], ftb=b[121])
        tras_min = self.txx_ns(mtb=_word(_lsn(b[27]), b[28]))
        trc_min  = self.txx_ns(mtb=_word(_msn(b[27]), b[29]), ftb=b[120])
        self.trfc = {
            "1x": (None, self.txx_ns(mtb=_word(b[31], b[30]))),
            "2x": (None, self.txx_ns(mtb=_word(b[33], b[32]))),
            "4x": (None, self.txx_ns(mtb=_word(b[35], b[34]))),
        }
        tfaw_min = self.txx_ns(mtb=_word(_lsn(b[36]), b[37]))
        trrd_s_min = self.txx_ns(mtb=b[38], ftb=b[119])
        trrd_l_min = self.txx_ns(mtb=b[39], ftb=b[118])
        tccd_l_min = self.txx_ns(mtb=b[40], ftb=b[117])  # min 6 cycles?
        twr_min  = self.txx_ns(mtb=_word(_lsn(b[41]), b[42]))
        twtr_s_min = self.txx_ns(mtb=_word(_lsn(b[43]), b[44]))
        twtr_l_min = self.txx_ns(mtb=_word(_msn(b[43]), b[45]))

        # minimum tFAW in clock cycles depends on page size
        sdram_device_width = {
            0b000:  4,
            0b001:  8,
            0b010: 16,
            0b011: 32,
        }[_read_field(b[12], nbits=3, shift=0)]
        page_size_bytes = self.ncols * sdram_device_width / 8
        tfaw_min_ck = {
            512:  16,
            1024: 20,
            2048: 28,
        }[page_size_bytes]

        technology_timings = _TechnologyTimings(
            tREFI = self.trefi,
            tWTR  = (4, twtr_l_min),
            tCCD  = (4, tccd_l_min),
            tRRD  = (4, trrd_l_min),
            tZQCS = (128, 80),
        )
        speedgrade_timings = _SpeedgradeTimings(
            tRP  = trp_min,
            tRCD = trcd_min,
            tWR  = twr_min,
            tRFC = self.trfc,
            tFAW = (tfaw_min_ck, tfaw_min),
            tRAS = tras_min,
        )

        self.speedgrade = str(self.speedgrade_freq(tckavg_min))
        self.technology_timings = technology_timings
        self.speedgrade_timings = {
            self.speedgrade: speedgrade_timings,
            "default": speedgrade_timings,
        }

def parse_spd_hexdump(filename):
    """Parse data dumped using the `spdread` command in LiteX BIOS

    This will read files in format:
        Memory dump:
        0x00000000  00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f  ................
        0x00000010  10 11 12 13 14 15 16 17 18 19 1a 1b 1c 1d 1e 1f  ................
    """
    data = []
    last_addr = -1
    with open(filename) as f:
        for line in f:
            if line.startswith("0x"):
                tokens = line.strip().split()
                addr = int(tokens[0], 16)
                assert addr > last_addr
                values = [int(v, 16) for v in tokens[1:17]]
                data.extend(values)
                last_addr = addr
    return data

# SDRAMModule --------------------------------------------------------------------------------------

Frac = namedtuple("Frac", ["num", "denom"])

class Timing(namedtuple("Timing", ["ck", "ns"])):
    def __add__(self, other):
        return Timing(self.ck + other.ck, self.ns + other.ns)

class SDRAMModule:
    """SDRAM module geometry and timings.

    SDRAM controller has to ensure that all geometry and
    timings parameters are fulfilled. Timings parameters
    can be expressed in ns, in SDRAM clock cycles or both
    and controller needs to use the greater value.

    SDRAM modules with the same geometry exist can have
    various speedgrades.
    """
    registered = False
    def __init__(self, clk_freq, rate, speedgrade=None, fine_refresh_mode=None):
        self.clk_freq      = clk_freq
        self.rate          = rate
        self.speedgrade    = speedgrade
        self.geom_settings = GeomSettings(
            bankbits = log2_int(self.nbanks),
            rowbits  = log2_int(self.nrows),
            colbits  = log2_int(self.ncols),
        )
        assert not (self.memtype != "DDR4" and fine_refresh_mode != None)
        assert fine_refresh_mode in [None, "1x", "2x", "4x"]
        if (fine_refresh_mode is None) and (self.memtype == "DDR4"):
            fine_refresh_mode = "1x"
        self.timing_settings = TimingSettings(
            tRP   = self.ck_ns_to_cycles(self.get("tRP")),
            tRCD  = self.ck_ns_to_cycles(self.get("tRCD")),
            tWR   = self.ck_ns_to_cycles(self.get("tWR")),
            tREFI = self.ck_ns_to_cycles(self.get("tREFI", fine_refresh_mode), margin=False),
            tRFC  = self.ck_ns_to_cycles(self.get("tRFC", fine_refresh_mode)),
            tWTR  = self.ck_ns_to_cycles(self.get("tWTR")),
            tFAW  = None if self.get("tFAW") is None else self.ck_ns_to_cycles(self.get("tFAW")),
            tCCD  = None if self.get("tCCD") is None else self.ck_ns_to_cycles(self.get("tCCD")),
            tRRD  = None if self.get("tRRD") is None else self.ck_ns_to_cycles(self.get("tRRD")),
            tRC   = None if self.get("tRAS") is None else self.ck_ns_to_cycles(self.get("tRP") + self.get("tRAS")),
            tRAS  = None if self.get("tRAS") is None else self.ck_ns_to_cycles(self.get("tRAS")),
            tZQCS = None if self.get("tZQCS") is None else self.ck_ns_to_cycles(self.get("tZQCS"))
        )
        self.timing_settings.fine_refresh_mode = fine_refresh_mode

    def get(self, name, key=None):
        assert name in _speedgrade_timings + _technology_timings, "Unknown name: {}".format(name)
        timing = self.get_timing(name)
        if timing is None:
            return None
        if (timing is not None) and (key is not None):
            timing = timing[key]
        if isinstance(timing, tuple):
            ck, ns = timing
        else:
            ck, ns = 0, timing
        ck = ck or 0
        ns = ns or 0
        return Timing(ck, ns)

    def get_timing(self, name):
        if name in _speedgrade_timings:
            return self.get_speedgrade_timing(name)
        return self.get_technology_timing(name)

    def get_speedgrade_timing(self, name):
        if hasattr(self, "speedgrade_timings"):
            speedgrade = "default" if self.speedgrade is None else self.speedgrade
            return getattr(self.speedgrade_timings[speedgrade], name)
        name = name + "_" + self.speedgrade if self.speedgrade is not None else name
        try:
            return getattr(self, name)
        except:
            pass

    def get_technology_timing(self, name):
        if hasattr(self, "technology_timings"):
            return getattr(self.technology_timings, name)
        else:
            try:
                return getattr(self, name)
            except:
                pass

    def ns_to_cycles(self, t, margin=True):
        clk_period_ns = 1e9/self.clk_freq
        t += self.margin if margin else 0
        return ceil(t/clk_period_ns)

    def ck_to_cycles(self, c):
        return ceil(c/self.rate_frac.denom)

    def ck_ns_to_cycles(self, timing, **kwargs):
        return max(self.ck_to_cycles(timing.ck), self.ns_to_cycles(timing.ns, **kwargs))

    @property
    def rate_frac(self):
        num, denom = map(int, self.rate.split(":"))
        assert num == 1, "Rate: numerator must be 1: {}".format(self.rate)
        return Frac(num, denom)

    @property
    def margin(self):
        clk_period_ns = 1e9 / self.clk_freq
        frac = self.rate_frac
        return clk_period_ns * (1 - frac.num/frac.denom)

    @classmethod
    def from_spd_data(cls, spd_data, clk_freq, fine_refresh_mode=None):
        # set parameters from SPD data based on memory type
        spd_cls = {
            0x0b: DDR3SPDData,
            0x0c: DDR4SPDData,
        }[spd_data[2]]
        spd = spd_cls(spd_data)

        # Create a deriving class to avoid modifying this one
        class _SDRAMModule(cls):
            memtype = spd.memtype
            nbanks = spd.nbanks
            nrows = spd.nrows
            ncols = spd.ncols
            technology_timings = spd.technology_timings
            speedgrade_timings = spd.speedgrade_timings
            # Save data for runtime verification
            _spd_data = spd_data

        nphases = {
            "SDR":   1,
            "DDR":   2,
            "LPDDR": 2,
            "DDR2":  2,
            "DDR3":  4,
            "DDR4":  4,
        }[spd.memtype]
        rate = "1:{}".format(nphases)

        return _SDRAMModule(clk_freq,
            rate              = rate,
            speedgrade        = spd.speedgrade,
            fine_refresh_mode = fine_refresh_mode)

class SDRAMRegisteredModule(SDRAMModule): registered = True

# SDR ----------------------------------------------------------------------------------------------

class SDRModule(SDRAMModule):                     memtype = "SDR"
class SDRRegisteredModule(SDRAMRegisteredModule): memtype = "SDR"

class IS42S16160(SDRModule):
    # geometry
    nbanks = 4
    nrows  = 8192
    ncols  = 512
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=20, tRCD=20, tWR=20, tRFC=(None, 70), tFAW=None, tRAS=None)}

class IS42S16320(SDRModule):
    # geometry
    nbanks = 4
    nrows  = 8192
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=20, tRCD=20, tWR=20, tRFC=(None, 70), tFAW=None, tRAS=None)}

class MT48LC4M16(SDRModule):
    # geometry
    nbanks = 4
    nrows  = 4096
    ncols  = 256
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=14, tRFC=(None, 66), tFAW=None, tRAS=None)}

class MT48LC16M16(SDRModule):
    # geometry
    nbanks = 4
    nrows  = 8192
    ncols  = 512
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=(None, 15))
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=20, tRCD=20, tWR=15, tRFC=(None, 66), tFAW=None, tRAS=44)}

class MT48LC32M8(SDRModule):
    # geometry
    nbanks = 4
    nrows  = 8192
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=(None, 15))
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=20, tRCD=20, tWR=15, tRFC=(None, 66), tFAW=None, tRAS=44)}

class AS4C4M16(SDRModule):
    # geometry
    nbanks = 4
    nrows  = 4096
    ncols  = 256
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/4096, tWTR=(2, None), tCCD=(1, None), tRRD=(None, 14))
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=22, tRCD=21, tWR=20, tRFC=(None, 63), tFAW=None, tRAS=42)}

class AS4C16M16(SDRModule):
    # geometry
    nbanks = 4
    nrows  = 8192
    ncols  = 512
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=18, tRCD=18, tWR=12, tRFC=(None, 60), tFAW=None, tRAS=None)}

class AS4C32M16(SDRModule):
    # geometry
    nbanks = 4
    nrows  = 8192
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=18, tRCD=18, tWR=12, tRFC=(None, 60), tFAW=None, tRAS=None)}

class AS4C32M8(SDRModule):
    # geometry
    nbanks = 4
    nrows  = 8192
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=(None, 15))
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=20, tRCD=20, tWR=15, tRFC=(None, 66), tFAW=None, tRAS=44)}

class M12L64322A(SDRModule):
    # geometry
    nbanks = 4
    nrows  = 2048
    ncols  = 256
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/4096, tWTR=(2, None), tCCD=(1, None), tRRD=(None, 10))
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 55), tFAW=None, tRAS=40)}

class M12L16161A(SDRModule):
    # geometry
    nbanks = 2
    nrows  = 2048
    ncols  = 256
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/4096, tWTR=(2, None), tCCD=(1, None), tRRD=(None, 10))
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 55), tFAW=None, tRAS=40)}

class W9825G6KH6(SDRModule):
    # geometry
    nbanks = 4
    nrows  = 8192
    ncols  = 512
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=(None, 10))
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 60), tFAW=None, tRAS=42)}

# DDR ----------------------------------------------------------------------------------------------

class DDRModule(SDRAMModule):                     memtype = "DDR"
class DDRRegisteredModule(SDRAMRegisteredModule): memtype = "DDR"

class MT46V32M16(DDRModule):
    # geometry
    nbanks = 4
    nrows  = 8192
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 70), tFAW=None, tRAS=None)}

# LPDDR --------------------------------------------------------------------------------------------

class LPDDRModule(SDRAMModule):                     memtype = "LPDDR"
class LPDDRRegisteredModule(SDRAMRegisteredModule): memtype = "LPDDR"

class MT46H32M16(LPDDRModule):
    # geometry
    nbanks = 4
    nrows  = 8192
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 72), tFAW=None, tRAS=None)}

class MT46H64M16(LPDDRModule):
    # geometry
    nbanks = 4
    nrows  = 16384
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 72), tFAW=None, tRAS=None)}


class MT46H32M32(LPDDRModule):
    # geometry
    nbanks = 4
    nrows  = 8192
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(2, None), tCCD=(1, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 72), tFAW=None, tRAS=None)}

# DDR2 ---------------------------------------------------------------------------------------------

class DDR2Module(SDRAMModule):                     memtype = "DDR2"
class DDR2RegisteredModule(SDRAMRegisteredModule): memtype = "DDR2"

class MT47H128M8(DDR2Module):
    # geometry
    nbanks = 8
    nrows  = 16384
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(None, 7.5), tCCD=(2, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 127.5), tFAW=None, tRAS=None)}

class MT47H32M16(DDR2Module):
    # geometry
    nbanks = 4
    nrows  = 8192
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(None, 7.5), tCCD=(2, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 127.5), tFAW=None, tRAS=None)}

class MT47H64M16(DDR2Module):
    # geometry
    nbanks = 8
    nrows  = 8192
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(None, 7.5), tCCD=(2, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(None, 127.5), tFAW=None, tRAS=None)}

class P3R1GE4JGF(DDR2Module):
    # geometry
    nbanks = 8
    nrows  = 8192
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(None, 7.5), tCCD=(2, None), tRRD=None)
    speedgrade_timings = {"default": _SpeedgradeTimings(tRP=12.5, tRCD=12.5, tWR=15, tRFC=(None, 127.5), tFAW=None, tRAS=None)}

# DDR3 (Chips) -------------------------------------------------------------------------------------

class DDR3Module(SDRAMModule):                     memtype = "DDR3"
class DDR3RegisteredModule(SDRAMRegisteredModule): memtype = "DDR3"

class AS4C128M16(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 16384
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6), tZQCS=(64, 80))
    speedgrade_timings = {
        "1600": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=13.75, tRFC=(160, None), tFAW=(None, 40), tRAS=35),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

class MT41K64M16(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 8192
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 10), tZQCS=(64, 80))
    speedgrade_timings = {
        "800":  _SpeedgradeTimings(tRP=13.1,  tRCD=13.1,  tWR=13.1,  tRFC=(64,  None), tFAW=(None, 50), tRAS=37.5),
        "1066": _SpeedgradeTimings(tRP=13.1,  tRCD=13.1,  tWR=13.1,  tRFC=(86,  None), tFAW=(None, 50), tRAS=37.5),
        "1333": _SpeedgradeTimings(tRP=13.5,  tRCD=13.5,  tWR=13.5,  tRFC=(107, None), tFAW=(None, 45), tRAS=36),
        "1600": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=13.75, tRFC=(128, None), tFAW=(None, 40), tRAS=35),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

class MT41J256M8(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 32768
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6), tZQCS=(64, 80))
    speedgrade_timings = {
        "800":  _SpeedgradeTimings(tRP=13.1,  tRCD=13.1,  tWR=15,  tRFC=(64, None),  tFAW=(None, 50), tRAS=37.5),
        "1066": _SpeedgradeTimings(tRP=13.1,  tRCD=13.1,  tWR=15,  tRFC=(86, None),  tFAW=(None, 50), tRAS=37.5),
        "1333": _SpeedgradeTimings(tRP=13.5,  tRCD=13.5,  tWR=15,  tRFC=(107, None), tFAW=(None, 45), tRAS=36),
    }
    speedgrade_timings["default"] = speedgrade_timings["1333"]

class MT41K256M8(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 32768
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6), tZQCS=(64, 80))
    speedgrade_timings = {
          "1600": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=15, tRFC=(128, None), tFAW=(None, 40), tRAS=35),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

class MT41J128M16(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 16384
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 10), tZQCS=(64, 80))
    speedgrade_timings = {
        "800":  _SpeedgradeTimings(tRP=13.1,  tRCD=13.1,  tWR=13.1,  tRFC=(64, None),  tFAW=(None, 50), tRAS=37.5),
        "1066": _SpeedgradeTimings(tRP=13.1,  tRCD=13.1,  tWR=13.1,  tRFC=(86, None),  tFAW=(None, 50), tRAS=37.5),
        "1333": _SpeedgradeTimings(tRP=13.5,  tRCD=13.5,  tWR=13.5,  tRFC=(107, None), tFAW=(None, 45), tRAS=36),
        "1600": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=13.75, tRFC=(128, None), tFAW=(None, 40), tRAS=35),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

class MT41K128M16(MT41J128M16): pass

class MT41J256M16(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 32768
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 10), tZQCS=(64, 80))
    speedgrade_timings = {
        "800":  _SpeedgradeTimings(tRP=13.1,  tRCD=13.1,  tWR=13.1,  tRFC=(139, None), tFAW=(None, 50), tRAS=37.5),
        "1066": _SpeedgradeTimings(tRP=13.1,  tRCD=13.1,  tWR=13.1,  tRFC=(138, None), tFAW=(None, 50), tRAS=37.5),
        "1333": _SpeedgradeTimings(tRP=13.5,  tRCD=13.5,  tWR=13.5,  tRFC=(174, None), tFAW=(None, 45), tRAS=36),
        "1600": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=13.75, tRFC=(208, None), tFAW=(None, 40), tRAS=35),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

class MT41K256M16(MT41J256M16): pass

class MT41J512M16(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 65536
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 10), tZQCS=(64, 80))
    speedgrade_timings = {
        "1600": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=13.75, tRFC=(280, None), tFAW=(None, 40), tRAS=39),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

class MT41K512M16(MT41J512M16): pass

class K4B1G0446F(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 16384
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 10), tZQCS=(64, 80))
    speedgrade_timings = {
        "800":  _SpeedgradeTimings(tRP=15,     tRCD=15,     tWR=15, tRFC=(120, None), tFAW=(None, 50), tRAS=37.5),
        "1066": _SpeedgradeTimings(tRP=13.125, tRCD=13.125, tWR=15, tRFC=(160, None), tFAW=(None, 50), tRAS=37.5),
        "1333": _SpeedgradeTimings(tRP=13.5,   tRCD=13.5,   tWR=15, tRFC=(200, None), tFAW=(None, 45), tRAS=36),
        "1600": _SpeedgradeTimings(tRP=13.75,  tRCD=13.75,  tWR=15, tRFC=(240, None), tFAW=(None, 40), tRAS=35),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

class K4B2G1646F(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 16384
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 10), tZQCS=(64, 80))
    speedgrade_timings = {
        "800":  _SpeedgradeTimings(tRP=15,     tRCD=15,     tWR=15, tRFC=(104, None), tFAW=(None, 50), tRAS=37.5),
        "1066": _SpeedgradeTimings(tRP=13.125, tRCD=13.125, tWR=15, tRFC=(139, None), tFAW=(None, 50), tRAS=37.5),
        "1333": _SpeedgradeTimings(tRP=13.5,   tRCD=13.5,   tWR=15, tRFC=(174, None), tFAW=(None, 45), tRAS=36),
        "1600": _SpeedgradeTimings(tRP=13.75,  tRCD=13.75,  tWR=15, tRFC=(208, None), tFAW=(None, 40), tRAS=35),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

class H5TC4G63CFR(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 16384
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 7.5), tZQCS=(64, 80))
    speedgrade_timings = {
        "800":  _SpeedgradeTimings(tRP=15, tRCD=15, tWR=15, tRFC=(260, None), tFAW=(None, 40), tRAS=37.5),
    }
    speedgrade_timings["default"] = speedgrade_timings["800"]

class IS43TR16128B(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 16384
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6), tZQCS=(64, 80))
    speedgrade_timings = {
        "1600": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=15, tRFC=(None, 160), tFAW=(None, 40), tRAS=35),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

class IS43TR16256A(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 32768
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6), tZQCS=(64, 80))
    speedgrade_timings = {
        "1600": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=15, tRFC=(None, 260), tFAW=(None, 30), tRAS=35),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

class IS43TR16512B(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 65536
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 10), tZQCS=(64, 80))
    speedgrade_timings = {
        "800":  _SpeedgradeTimings(tRP=15,     tRCD=15,     tWR=15, tRFC=(None, 350), tFAW=(None, 50), tRAS=37.5),
        "1066": _SpeedgradeTimings(tRP=13.125, tRCD=13.125, tWR=15, tRFC=(None, 350), tFAW=(None, 50), tRAS=37.5),
        "1333": _SpeedgradeTimings(tRP=13.5,   tRCD=13.5,   tWR=15, tRFC=(None, 350), tFAW=(None, 50), tRAS=36),
        "1600": _SpeedgradeTimings(tRP=13.75,  tRCD=13.75,  tWR=15, tRFC=(None, 350), tFAW=(None, 50), tRAS=35),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

# DDR3 (SO-DIMM) -----------------------------------------------------------------------------------

class MT8JTF12864(DDR3Module):
    # base chip: MT41J128M8
    # geometry
    nbanks = 8
    nrows  = 16384
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6), tZQCS=(64, 80))
    speedgrade_timings = {
        "1066": _SpeedgradeTimings(tRP=15,     tRCD=15,     tWR=15, tRFC=(None, 110), tFAW=(None, 37.5), tRAS=37.5),
        "1333": _SpeedgradeTimings(tRP=13.125, tRCD=13.125, tWR=15, tRFC=(None, 110), tFAW=(None, 30),   tRAS=36),
    }
    speedgrade_timings["default"] = speedgrade_timings["1333"]

class MT8KTF51264(DDR3Module):
    # base chip: MT41K512M8
    # geometry
    nbanks = 8
    nrows  = 65536
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6), tZQCS=(64, 80))
    speedgrade_timings = {
        "800" : _SpeedgradeTimings(tRP=15,     tRCD=15,     tWR=15, tRFC=(None, 260), tFAW=(None, 40), tRAS=37.5),
        "1066": _SpeedgradeTimings(tRP=15,     tRCD=15,     tWR=15, tRFC=(None, 260), tFAW=(None, 40), tRAS=37.5),
        "1333": _SpeedgradeTimings(tRP=13.125, tRCD=13.125, tWR=15, tRFC=(None, 260), tFAW=(None, 30), tRAS=36),
        "1600": _SpeedgradeTimings(tRP=13.125, tRCD=13.125, tWR=15, tRFC=(None, 260), tFAW=(None, 30), tRAS=35),
        "1866": _SpeedgradeTimings(tRP=13.125, tRCD=13.125, tWR=15, tRFC=(None, 260), tFAW=(None, 27), tRAS=34),
    }
    speedgrade_timings["default"] = speedgrade_timings["1866"]

class MT18KSF1G72HZ(DDR3Module):
    # base chip: MT41K512M8
    # geometry
    nbanks = 8
    nrows  = 65536
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6), tZQCS=(64, 80))
    speedgrade_timings = {
        "1066": _SpeedgradeTimings(tRP=15,     tRCD=15,     tWR=15, tRFC=(None, 260), tFAW=(None, 40), tRAS=37.5),
        "1333": _SpeedgradeTimings(tRP=13.125, tRCD=13.125, tWR=15, tRFC=(None, 260), tFAW=(None, 30), tRAS=36),
        "1600": _SpeedgradeTimings(tRP=13.125, tRCD=13.125, tWR=15, tRFC=(None, 260), tFAW=(None, 30), tRAS=35),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

class AS4C256M16D3A(DDR3Module):
    # geometry
    nbanks = 8
    nrows  = 32768
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 7.5), tZQCS=(64, 80))
    speedgrade_timings = {
        "1600": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=15, tRFC=(None, 260), tFAW=(None, 40), tRAS=35),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

class MT16KTF1G64HZ(DDR3Module):
    # base chip: MT41K512M8
    # geometry
    nbanks = 8
    nrows  = 65536
    ncols  = 1024
    # timings
    technology_timings = _TechnologyTimings(tREFI=64e6/8192, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6), tZQCS=(64, 80))
    speedgrade_timings = {
        "800" : _SpeedgradeTimings(tRP=15,     tRCD=15,     tWR=15, tRFC=(None, 260), tFAW=(None, 40), tRAS=37.5),
        "1066": _SpeedgradeTimings(tRP=15,     tRCD=15,     tWR=15, tRFC=(None, 260), tFAW=(None, 40), tRAS=37.5),
        "1333": _SpeedgradeTimings(tRP=15,     tRCD=15,     tWR=15, tRFC=(None, 260), tFAW=(None, 30), tRAS=36),
        "1600": _SpeedgradeTimings(tRP=13.125, tRCD=13.125, tWR=15, tRFC=(None, 260), tFAW=(None, 30), tRAS=35),
        "1866": _SpeedgradeTimings(tRP=13.125, tRCD=13.125, tWR=15, tRFC=(None, 260), tFAW=(None, 27), tRAS=34),
    }
    speedgrade_timings["default"] = speedgrade_timings["1866"]

# RPC ----------------------------------------------------------------------------------------------

class RPCModule(SDRAMModule): memtype = "RPC"

class EM6GA16L(RPCModule):
    # 64MBits per bank => 256Mb
    # TODO: the timings may need further verification
    # geometry
    nbanks = 4     #
    ncols = 1024
    nrows = 4096  # 64M / 1024 / 16 = 4k
    # timings
    technology_timings = _TechnologyTimings(
        tREFI=64e6/8192,
        # It seems that we need to calculate tCCD from tBESL by assuming BC=1, then it is:
        # RL=WL + burst_time + tBESL
        tCCD=(13+1 + 8 + 7, None),
        # CCD enforces tBESL for read commands, we use tWTR to ensure it for writes
        tWTR=(11 + 2, None),  # 11 for tBESL + 2 for STB low before a command
        tRRD=(None, 7.5),
        tZQCS=(None, 90),
    )
    speedgrade_timings = {
        # FIXME: we're increasing tWR by 1 sysclk to compensate for long write (assuming max sysclk 200 MHz)
        # Should we use tRFQSd for tRFC?
        "1600": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=15 + 1e9/200e6, tRFC=(3*100, None), tFAW=None, tRAS=35),
        "1866": _SpeedgradeTimings(tRP=13.91, tRCD=13.91, tWR=15 + 1e9/200e6, tRFC=(3*100, None), tFAW=None, tRAS=34),
    }
    speedgrade_timings["default"] = speedgrade_timings["1600"]

# DDR4 (Chips) -------------------------------------------------------------------------------------

class DDR4Module(SDRAMModule):                     memtype = "DDR4"
class DDR4RegisteredModule(SDRAMRegisteredModule): memtype = "DDR4"

class EDY4016A(DDR4Module):
    # geometry
    ngroupbanks = 4
    ngroups     = 2
    nbanks      = ngroups * ngroupbanks
    nrows       = 32768
    ncols       = 1024
    # timings
    trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
    trfc  = {"1x": (None, 260), "2x": (None, 160),   "4x": (None, 110)}
    technology_timings = _TechnologyTimings(tREFI=trefi, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 4.9), tZQCS=(128, 80))
    speedgrade_timings = {
        "2400": _SpeedgradeTimings(tRP=13.32, tRCD=13.32, tWR=15, tRFC=trfc, tFAW=(28, 30), tRAS=32),
    }
    speedgrade_timings["default"] = speedgrade_timings["2400"]

class MT40A1G8(DDR4Module):
    # geometry
    ngroupbanks = 4
    ngroups     = 4
    nbanks      = ngroups * ngroupbanks
    nrows       = 65536
    ncols       = 1024
    # timings
    trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
    trfc  = {"1x": (None, 350), "2x": (None, 260),   "4x": (None, 160)}
    technology_timings = _TechnologyTimings(tREFI=trefi, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6.4), tZQCS=(128, 80))
    speedgrade_timings = {
        "2400": _SpeedgradeTimings(tRP=13.32, tRCD=13.32, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=32),
        "2666": _SpeedgradeTimings(tRP=13.50, tRCD=13.50, tWR=15, tRFC=trfc, tFAW=(20, 21), tRAS=32),
    }
    speedgrade_timings["default"] = speedgrade_timings["2400"]

class MT40A2G8(DDR4Module):
    # geometry
    ngroupbanks = 4
    ngroups     = 4
    nbanks      = ngroups * ngroupbanks
    nrows       = 131072
    ncols       = 1024
    # timings
    trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
    trfc  = {"1x": (None, 350), "2x": (None, 260),   "4x": (None, 160)}
    technology_timings = _TechnologyTimings(tREFI=trefi, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 6.4), tZQCS=(128, 80))
    speedgrade_timings = {
        "2400": _SpeedgradeTimings(tRP=13.32, tRCD=13.32, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=32),
        "2666": _SpeedgradeTimings(tRP=13.50, tRCD=13.50, tWR=15, tRFC=trfc, tFAW=(20, 21), tRAS=32),
    }
    speedgrade_timings["default"] = speedgrade_timings["2400"]

class MT40A2G16(MT40A2G8):
    pass # TwinDie MT40A2G8.

class MT40A256M16(DDR4Module):
    # geometry
    ngroupbanks = 4
    ngroups     = 2
    nbanks      = ngroups * ngroupbanks
    nrows       = 32768
    ncols       = 1024
    # timings
    trefi = {"1x": 64e6/8192, "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
    trfc  = {"1x": (None, 260), "2x": (None, 160), "4x": (None, 110)}
    technology_timings = _TechnologyTimings(tREFI=trefi, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 4.9), tZQCS=(128, 80))
    speedgrade_timings = {
        "2400": _SpeedgradeTimings(tRP=13.32, tRCD=13.32, tWR=15, tRFC=trfc, tFAW=(28, 35), tRAS=32),
    }
    speedgrade_timings["default"] = speedgrade_timings["2400"]

class MT40A512M8(DDR4Module):
    # geometry
    ngroupbanks = 4
    ngroups     = 4
    nbanks      = ngroups * ngroupbanks
    nrows       = 32768
    ncols       = 1024
    # timings
    trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
    trfc  = {"1x": (None, 350), "2x": (None, 260),   "4x": (None, 160)}
    technology_timings = _TechnologyTimings(tREFI=trefi, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 4.9), tZQCS=(128, 80))
    speedgrade_timings = {
        "2400": _SpeedgradeTimings(tRP=13.32, tRCD=13.32, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=32),
        "2666": _SpeedgradeTimings(tRP=13.50, tRCD=13.50, tWR=15, tRFC=trfc, tFAW=(20, 21), tRAS=32),
    }
    speedgrade_timings["default"] = speedgrade_timings["2400"]

class MT40A512M16(DDR4Module):
    # geometry
    ngroupbanks = 4
    ngroups     = 2
    nbanks      = ngroups * ngroupbanks
    nrows       = 65536
    ncols       = 1024
    # timings
    trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
    trfc  = {"1x": (None, 350), "2x": (None, 260),   "4x": (None, 160)}
    technology_timings = _TechnologyTimings(tREFI=trefi, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 4.9), tZQCS=(128, 80))
    speedgrade_timings = {
        "2400": _SpeedgradeTimings(tRP=13.32, tRCD=13.32, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=32),
    }
    speedgrade_timings["default"] = speedgrade_timings["2400"]

# DDR4 (SO-DIMM) -----------------------------------------------------------------------------------

class KVR21SE15S84(DDR4Module):
    # geometry
    ngroupbanks = 4
    ngroups     = 4
    nbanks      = ngroups * ngroupbanks
    nrows       = 32768
    ncols       = 1024
    # timings
    trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
    trfc  = {"1x": (None, 350), "2x": (None, 260),   "4x": (None, 160)}
    technology_timings = _TechnologyTimings(tREFI=trefi, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 4.9), tZQCS=(128, 80))
    speedgrade_timings = {
        "2133": _SpeedgradeTimings(tRP=13.5, tRCD=13.5, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=33),
    }
    speedgrade_timings["default"] = speedgrade_timings["2133"]

class MTA4ATF51264HZ(DDR4Module):
    # geometry
    ngroupbanks = 4
    ngroups     = 2
    nbanks      = ngroups * ngroupbanks
    nrows       = 65536
    ncols       = 1024
    # timings
    trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
    trfc  = {"1x": (None, 350), "2x": (None, 260),   "4x": (None, 160)}
    technology_timings = _TechnologyTimings(tREFI=trefi, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 4.9), tZQCS=(128, 80))
    speedgrade_timings = {
        "2133": _SpeedgradeTimings(tRP=13.5, tRCD=13.5, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=33),
    }
    speedgrade_timings["default"] = speedgrade_timings["2133"]

# DDR4 (RDIMM) -------------------------------------------------------------------------------------

class MTA18ASF2G72PZ(DDR4RegisteredModule):
    # geometry
    ngroupbanks = 4
    ngroups     = 4
    nbanks      = ngroups * ngroupbanks
    nrows       = 131072
    ncols       = 1024
    # timings
    trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
    trfc  = {"1x": (None, 350), "2x": (None, 260),   "4x": (None, 160)}
    technology_timings = _TechnologyTimings(tREFI=trefi, tWTR=(4, 7.5), tCCD=(4, None), tRRD=(4, 4.9), tZQCS=(128, 80))
    speedgrade_timings = {
        "2400": _SpeedgradeTimings(tRP=14.16, tRCD=14.16, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=32),
        "2666": _SpeedgradeTimings(tRP=14.25, tRCD=14.25, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=32),
        "2933": _SpeedgradeTimings(tRP=14.32, tRCD=14.32, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=32),
        "3200": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=15, tRFC=trfc, tFAW=(20, 25), tRAS=32),
    }
    speedgrade_timings["default"] = speedgrade_timings["2400"]

class MTA36ASF4G72PZ(MTA18ASF2G72PZ): pass
    # This module is similar to MTA18ASF2G72PZ, with the exception that it is a dual-rank module. The rank number
    # can be specified in the PHY settings.

class HMA82GR7DJR4N(DDR4RegisteredModule):
    # geometry
    ngroupbanks = 4
    ngroups     = 4
    nbanks      = ngroups * ngroupbanks
    nrows       = 131072
    ncols       = 1024
    # timings
    trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
    trfc  = {"1x": (None, 350), "2x": (None, 260),   "4x": (None, 160)}
    technology_timings = _TechnologyTimings(tREFI=trefi, tWTR=(4, 7.5), tCCD=(4, 5), tRRD=(4, 4.9), tZQCS=(128, 80))
    speedgrade_timings = {
        "3200": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=15, tRFC=trfc, tFAW=10, tRAS=32),
    }
    speedgrade_timings["default"] = speedgrade_timings["3200"]

class HMA84GR7DJR4N(HMA82GR7DJR4N): pass
    # This module is similar to HMA82GR7DJR4N, with the exception that it is a dual-rank module. The rank number
    # can be specified in the PHY settings.

class M393A2K40DB3(DDR4RegisteredModule):
    # geometry
    ngroupbanks = 4
    ngroups     = 4
    nbanks      = ngroups * ngroupbanks
    nrows       = 131072
    ncols       = 1024
    # timings
    trefi = {"1x": 64e6/8192,   "2x": (64e6/8192)/2, "4x": (64e6/8192)/4}
    trfc  = {"1x": (None, 350), "2x": (None, 260),   "4x": (None, 160)}
    technology_timings = _TechnologyTimings(tREFI=trefi, tWTR=(4, 7.5), tCCD=(4, 5), tRRD=(4, 4.9), tZQCS=(128, 80))
    speedgrade_timings = {
        "3200": _SpeedgradeTimings(tRP=13.75, tRCD=13.75, tWR=15, tRFC=trfc, tFAW=10, tRAS=32),
    }
    speedgrade_timings["default"] = speedgrade_timings["3200"]

class M393A4K40DB3(M393A2K40DB3): pass
    # This module is similar to M393A2K40DB3, with the exception that it is a dual-rank module. The rank number
    # can be specified in the PHY settings.

# LPDDR4 -------------------------------------------------------------------------------------------

class MT53E256M16D1(SDRAMModule):
    memtype = "LPDDR4"

    nbanks = 8
    nrows = 32768
    ncols = 1024

    # TODO: find a way to select if we need masked writes
    tccd = {"write": (8, None), "masked-write": (32, None)}

    # TODO: tZQCS - performing ZQC during runtime will require modifying Refresher, as ZQC has to be done in 2 phases
    # 1. ZQCAL START is issued 2. ZQCAL LATCH updates the values, the time START->LATCH tZQCAL=1us, so we cannot block
    # the controller during this time, after ZQCAL LATCH we have to wait tZQLAT=max(8ck, 30ns)
    technology_timings = _TechnologyTimings(tREFI=32e6/8192, tWTR=(8, 10), tCCD=tccd["masked-write"], tRRD=(4, 10), tZQCS=None)
    speedgrade_timings = {
        "1866": _SpeedgradeTimings(tRP=(3, 21), tRCD=(4, 18), tWR=(4, 18), tRFC=180, tFAW=40, tRAS=(3, 42)),  # TODO: tRAS_max
    }
    speedgrade_timings["default"] = speedgrade_timings["1866"]



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

import re
import math
from fractions import Fraction
from functools import reduce
from operator import or_
from collections import defaultdict

from migen import *

from litex.soc.interconnect import stream

from litedram.common import TappedDelayLine


def bit(n, val):
    return (val & (1 << n)) >> n

def chunks(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i:i + n]

def bitpattern(s):
    if len(s) > 8:
        return reduce(or_, [bitpattern(si) << (8*i) for i, si in enumerate(chunks(s, 8))])
    assert len(s) == 8
    s = s.translate(s.maketrans("_-", "01"))
    return int(s[::-1], 2)  # LSB first, so reverse the string

def delayed(mod, sig, cycles=1, **kwargs):
    delay = TappedDelayLine(signal=sig, ntaps=cycles, **kwargs)
    mod.submodules += delay
    return delay.output

def edge(mod, cond):
    """Get a signal that is high on a rising edge of `cond`"""
    cond_d = Signal()
    mod.sync += cond_d.eq(cond)
    return  ~cond_d & cond


class ConstBitSlip(Module):
    def __init__(self, dw, slp, cycles, i=None, o=None, register=True):
        self.i = Signal(dw, name='i') if i is None else i
        self.o = Signal(dw, name='o') if o is None else o

        assert cycles >= 1, cycles
        assert 0 <= slp <= cycles*dw-1, (slp, cycles, dw)
        slp = (cycles*dw-1) - slp

        # # #

        self.r = r = Signal((cycles+1)*dw, reset_less=True)
        if register:
            self.sync += r.eq(Cat(r[dw:], self.i))
        else:
            reg = Signal(cycles*dw, reset_less=True)
            # Cat with slice of len=0 generates incorrect Verilog
            if len(reg[dw:]) > 0:
                self.sync += reg.eq(Cat(reg[dw:], self.i))
            else:
                self.sync += reg.eq(self.i)
            self.comb += r.eq(Cat(reg, self.i))
        self.comb += self.o.eq(r[slp+1:dw+slp+1])

    @staticmethod
    def min_cycles(slp, dw):
        """Minimum number of cycles to be able to use given bitslip values"""
        return math.ceil((slp + 1) / dw)


# TODO: rewrite DQSPattern in litedram/common.py to support different data widths
class DQSPattern(Module):
    def __init__(self, preamble=None, postamble=None, wlevel_en=0, wlevel_strobe=0, register=False):
        self.preamble  = Signal() if preamble  is None else preamble
        self.postamble = Signal() if postamble is None else postamble
        self.o = Signal(16)

        # # #

        # DQS Pattern transmitted as LSB-first.

        self.comb += [
            self.o.eq(0b0101010101010101),
            If(self.preamble,
                self.o.eq(0b0101000001010101)  # 2tCK write preamble
            ),
            If(self.postamble,
                self.o.eq(0b0101010101010100)
            ),
            If(wlevel_en,
                self.o.eq(0b0000000000000000),
                If(wlevel_strobe,
                    # use 2 toggles as, according to datasheet, the first one may not be registered
                    self.o.eq(0b0000000000000101)
                )
            )
        ]
        if register:
            o = Signal.like(self.o)
            self.sync += o.eq(self.o)
            self.o = o


class Latency:
    """Helper for specifying latency in different clock domains"""

    PATTERN = re.compile(r"^sys(?:(\d+)x)?$")

    def __init__(self, **kwargs):
        self._sys = Fraction(0, 1)
        for name, cycles in kwargs.items():
            m = self.PATTERN.match(name)
            assert m, f"Wrong format: {name}"
            denom = m.group(1) or 1
            self._sys += Fraction(cycles, int(denom))

    def __getattr__(self, name):
        m = self.PATTERN.match(name)
        if m:
            denom = m.group(1) or 1
            cycles = self._sys * int(denom)
            if cycles.denominator != 1:
                raise ValueError("{}.{} results in a fraction: {}".format(self, name, cycles))
            return cycles.numerator
        raise AttributeError(name)

    def __add__(self, other):
        new = Latency()
        new._sys = self._sys + other._sys
        return new

    def __repr__(self):
        return "Latency({} sys clk)".format(self._sys)


class CommandsPipeline(Module):
    """Commands pipeline logic for LPDDR4/LPDDR5

    Single DFI command may require more than one LPDDR4/LPDDR5 command, effectively
    spanning multiple DFI phases. This module given a list of DFI phase adapters
    will use them to translate DFI commands to data on the CS/CA lines and will
    handle any possible overlaps.

    Basic check will make sure that no command will be sent to DRAM if there was any command
    sent by the controller on DFI during previous phases. The extended version will instead
    make sure no command is sent to DRAM if there was any command _actually sent to DRAM_
    during the previous phasess. This is more expensive in terms of resources and generally
    not needed.

    Adapters have to provide the following fields: valid, cs, ca.
    """
    def __init__(self, adapters, *,
            cs_ser_width,  # n bits serialized in controller cycle (depends on CS being SDR/DDR)
            ca_ser_width,  # n bits serialized in controller cycle (depends on CA being SDR/DDR)
            ca_nbits,      # number of CA lines (LPDDR4/5 -> 6/7)
            cmd_nphases_span,  # at most how many phases can a command span
            extended_overlaps_check=False):
        nphases = len(adapters)
        self.cs = Signal(cs_ser_width)
        self.ca = [Signal(ca_ser_width) for _ in range(ca_nbits)]
        assert cmd_nphases_span <= nphases

        # # #

        # Number of phases (before the current one) we need to check for overlaps
        n_previous = cmd_nphases_span - 1
        # Number of bits to slip CA per phase (how many CA output bits are equivalent to 1 CS output bit)
        assert ca_ser_width % cs_ser_width == 0, f"Non-integer CA:CS output width ratio: {ca_ser_width % cs_ser_width}"
        ca_phase_slip = ca_ser_width // cs_ser_width

        # Create a history of valid adapters used for masking overlapping ones
        valids = ConstBitSlip(dw=nphases, slp=0, cycles=1, register=False)
        self.submodules += valids
        self.comb += valids.i.eq(Cat(a.valid for a in adapters))
        valids_hist = valids.r
        if extended_overlaps_check:
            valids_hist = Signal.like(valids.r)
            for i in range(len(valids_hist)):
                hist_before = valids_hist[max(0, i-n_previous):i]
                was_valid_before = reduce(or_, hist_before, 0)
                self.comb += valids_hist[i].eq(valids.r[i] & ~was_valid_before)

        cs_per_adapter = []
        ca_per_adapter = defaultdict(list)
        for phase, adapter in enumerate(adapters):
            # The signals from an adapter can be used if there were no commands on previous cycles
            allowed = ~reduce(or_, valids_hist[nphases+phase - n_previous:nphases+phase])

            # Use CS and CA of given adapter slipped by `phase` bits
            cs_bs = ConstBitSlip(dw=cs_ser_width, slp=phase, cycles=1)
            self.submodules += cs_bs
            cs_mask = Replicate(allowed, len(cs_bs.i))
            self.comb += cs_bs.i.eq(Cat(adapter.cs) & cs_mask),
            cs_per_adapter.append(cs_bs.o)

            # For CA we need to do the same for each bit
            ca_bits = []
            for bit in range(ca_nbits):
                ca_bs = ConstBitSlip(dw=ca_ser_width, slp=phase*ca_phase_slip, cycles=1)
                self.submodules += ca_bs
                ca_bit_hist = [ca[bit] for ca in adapter.ca]
                ca_mask = Replicate(allowed, len(ca_bs.o))
                self.comb += ca_bs.i.eq(Cat(*ca_bit_hist) & ca_mask),
                ca_per_adapter[bit].append(ca_bs.o)

        # OR all the masked signals
        self.comb += self.cs.eq(reduce(or_, cs_per_adapter))
        for bit in range(ca_nbits):
            self.comb += self.ca[bit].eq(reduce(or_, ca_per_adapter[bit]))


class Serializer(Module):
    """Serialize given input signal

    The parallel part uses `clkdiv` to latch the data. Output data counter `cnt` is incremented
    on rising edges of `clk` and it determines current slice of `i` that is presented on `o`.
    `LATENCY` is specified in `clkdiv` cycles.

    NOTE: both `clk` and `clkdiv` should be phase aligned.
    NOTE: `reset_cnt` is set to `ratio - 1` so that on the first clock edge after reset it is 0
    """
    # TODO: make dynamic (0 for register=False)
    LATENCY = 1

    def __init__(self, clkdiv, clk, i_dw, o_dw, i=None, o=None, reset=None, register=True,
            reset_cnt=-1, name=None):
        assert i_dw > o_dw, (i_dw, o_dw)
        assert i_dw % o_dw == 0, (i_dw, o_dw)
        ratio = i_dw // o_dw

        sd_clk = getattr(self.sync, clk)
        sd_clkdiv = getattr(self.sync, clkdiv)

        if i is None: i = Signal(i_dw)
        if o is None: o = Signal(o_dw)
        if reset is None: reset = Signal()

        self.i = i
        self.o = o
        self.reset = reset

        if reset_cnt < 0:
            reset_cnt = ratio + reset_cnt

        # Serial part
        cnt = Signal(max=ratio, reset=reset_cnt, name='{}_cnt'.format(name) if name is not None else None)
        sd_clk += If(reset | cnt == ratio - 1, cnt.eq(0)).Else(cnt.eq(cnt + 1))

        # Parallel part
        if register:
            i_d = Signal.like(self.i)
            sd_clkdiv += i_d.eq(self.i)
            i = i_d
        i_array = Array([i[n*o_dw:(n+1)*o_dw] for n in range(ratio)])
        self.comb += self.o.eq(i_array[cnt])


class Deserializer(Module):
    """Deserialize given input signal

    The serial part latches the input data on rising edges of `clk` and stores them in the `o_pre`
    buffer. The parallel part presents the data on `clkdiv` rising edges. `LATENCY` is expressed in
    `clkdiv` cycles. The additional latency cycle (compared to Serializer) is used to ensure that
    the last input bit is deserialized correctly.

    NOTE: both `clk` and `clkdiv` should be phase aligned.
    NOTE: `reset_cnt` is set to `ratio - 1` so that on the first clock edge after reset it is 0
    """
    LATENCY = 2

    def __init__(self, clkdiv, clk, i_dw, o_dw, i=None, o=None, reset=None, reset_cnt=-1, name=None):
        assert i_dw < o_dw, (i_dw, o_dw)
        assert o_dw % i_dw == 0, (i_dw, o_dw)
        ratio = o_dw // i_dw

        sd_clk = getattr(self.sync, clk)
        sd_clkdiv = getattr(self.sync, clkdiv)

        if i is None: i = Signal(i_dw)
        if o is None: o = Signal(o_dw)
        if reset is None: reset = Signal()

        self.i = i
        self.o = o
        self.reset = reset

        if reset_cnt < 0:
            reset_cnt = ratio + reset_cnt

        # Serial part
        cnt = Signal(max=ratio, reset=reset_cnt, name='{}_cnt'.format(name) if name is not None else None)
        sd_clk += If(reset, cnt.eq(0)).Else(cnt.eq(cnt + 1))

        def as_array(out):
            return Array([out[n*i_dw:(n+1)*i_dw] for n in range(ratio)])

        o_pre = Signal.like(self.o)
        sd_clk += as_array(o_pre)[cnt].eq(self.i)

        # Parallel part
        # we need to ensure that the last chunk will be correct if clocks are phase aligned
        o_pre_d = Signal.like(self.o)
        sd_clkdiv += o_pre_d.eq(o_pre)
        # would work as self.comb (at least in simulation)
        sd_clkdiv += self.o.eq(Cat(as_array(o_pre_d)[:-1], as_array(o_pre)[-1]))


class HoldValid(Module):
    """Hold input data until ready

    Acts more or less like PipeValid with 0 latency. Data on source becomes valid in the same
    cycle on which it is valid on sink. In the next cycle the data is latched to a buffer and
    the data from buffer is presented on source. After source.ready, it resets back to sink data.
    """
    def __init__(self, layout):
        self.sink = stream.Endpoint(layout)
        self.buf = stream.Endpoint(layout)
        self.source = stream.Endpoint(layout)

        self.sync += [
            If(self.buf.ready,
                self.buf.valid.eq(0),
            ).Elif(self.sink.valid,
                self.buf.valid.eq(1),
                self.buf.payload.eq(self.sink.payload),
                self.buf.param.eq(self.sink.param),
            ),
        ]

        self.comb += [
            If(self.buf.valid,
                self.buf.connect(self.source),
            ).Else(
                self.sink.connect(self.source),
            ),
        ]



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2021 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.soc.interconnect.csr import CSR

from litedram.phy.utils import delayed, Serializer, Deserializer, Latency, ConstBitSlip
from litedram.phy.sim_utils import SimPad, SimulationPads, SimSerDesMixin
from litedram.phy.lpddr5.basephy import LPDDR5PHY


class LPDDR5SimulationPads(SimulationPads):
    def layout(self, databits=16):
        return [
            SimPad("reset_n", 1),
            SimPad("ck", 1),
            SimPad("cs", 1),
            SimPad("ca", 7),
            SimPad("dq", databits, io=True),
            SimPad("wck", databits//8),
            SimPad("rdqs", databits//8, io=True),
            SimPad("dmi", databits//8, io=True),
        ]


class LPDDR5SimPHY(SimSerDesMixin, LPDDR5PHY):
    """LPDDR5 simulation PHY

    The following clock domains are used:
        * MC @ sys
        * PHY CK @ sys
        * PHY WCK @ sys2x or sys4x

    For simulation purpose two additional "DDR" clock domains are required (sys4x or sys8x
    depending on WCK ratio).
    """
    def __init__(self, sys_clk_freq, wck_ck_ratio=2, aligned_reset_zero=False, **kwargs):
        pads = LPDDR5SimulationPads()
        self.submodules += pads
        super().__init__(pads,
            ck_freq      = sys_clk_freq,
            wck_ck_ratio = wck_ck_ratio,
            ser_latency  = Latency(sys=Serializer.LATENCY),
            des_latency  = Latency(sys=Deserializer.LATENCY),
            phytype      = "LPDDR5SimPHY",
            **kwargs)

        # fake delays (make no nsense in simulation, but sdram.c expects them)
        self.settings.read_leveling = True
        self._rdly_dq_rst = CSR()
        self._rdly_dq_inc = CSR()

        self.settings.write_leveling = True
        self._cdly_rst = CSR()
        self._cdly_inc = CSR()
        self._wdly_dq_rst = CSR()
        self._wdly_dq_inc = CSR()
        self._wdly_dqs_rst = CSR()
        self._wdly_dqs_inc = CSR()
        self._half_sys8x_taps = CSR()

        self.settings.delays = 1

        delay = lambda sig, cycles: delayed(self, sig, cycles=cycles)
        ddr_ck      = dict(clkdiv="sys", clk="sys2x")
        ddr_ca      = dict(clkdiv="sys", clk="sys4x")
        ddr_wck     = dict(clkdiv="sys", clk={2: "sys4x", 4: "sys8x"}[wck_ck_ratio])
        ddr_wck_180 = dict(clkdiv="sys", clk={2: "sys4x_180", 4: "sys8x_180"}[wck_ck_ratio])

        def cdc(sig, cd):
            latched = Signal.like(sig)
            sd_wck = getattr(self.sync, cd["clk"])
            sd_wck += latched.eq(delay(sig, cycles=Serializer.LATENCY))
            return dict(i=latched, register=False)

        if aligned_reset_zero:
            ddr_ck["reset_cnt"] = 0
            ddr_ca["reset_cnt"] = 0
            ddr_wck["reset_cnt"] = 0

        self.comb += self.pads.reset_n.eq(delay(self.out.reset_n, cycles=Serializer.LATENCY))

        # CK signals
        self.ser(i=self.out.ck, o=self.pads.ck, name='ck', **ddr_ck)

        # CS (SDR) is delayed by 180 deg to be center aligned with CK
        # Use ConstBitSlip to shift it, then serialize that 2-bit signal as DDR (like with CK)
        cs_2bit = Signal(2)
        cs_2bit_d = Signal(2)
        self.comb += cs_2bit.eq(Replicate(self.out.cs, 2))
        self.submodules += ConstBitSlip(dw=2, slp=1, cycles=1, register=False, i=cs_2bit, o=cs_2bit_d)
        self.ser(i=cs_2bit_d, o=self.pads.cs, name='cs', **ddr_ck)

        # To center align CA (DDR) with CK it has to be delayed by 270 deg CK (+90 deg relative to CS)
        for i in range(7):
            # To achieve 270 deg shift we use ConstBitSlip with slp=3 and dw=4.
            # For this to work we first widen CA to 4 bits and use sys4x when serializing.
            ca_4bit = Signal(4)
            ca_4bit_d = Signal(4)
            self.comb += ca_4bit.eq(Cat([Replicate(bit, 2) for bit in self.out.ca[i]]))
            self.submodules += ConstBitSlip(dw=4, slp=3, cycles=1, register=False, i=ca_4bit, o=ca_4bit_d)
            self.ser(i=ca_4bit_d, o=self.pads.ca[i], name=f'ca{i}', **ddr_ca)

        # WCK
        for i in range(self.databits//8):
            self.ser(i=self.out.wck[i], o=self.pads.wck[i], name=f'wck{i}', **ddr_wck_180)
            self.ser(i=self.out.dmi_o[i], o=self.pads.dmi_o[i], name=f'dmi_o{i}', **ddr_wck)
            self.des(o=self.out.dmi_i[i], i=self.pads.dmi[i],   name=f'dmi_i{i}', **ddr_wck)
            self.ser(i=self.out.rdqs_o[i], o=self.pads.rdqs_o[i], name=f'rdqs_o{i}', **ddr_wck_180)
            self.des(o=self.out.rdqs_i[i], i=self.pads.rdqs[i],   name=f'rdqs_i{i}', **ddr_wck_180)
        for i in range(self.databits):
            self.ser(i=self.out.dq_o[i], o=self.pads.dq_o[i], name=f'dq_o{i}', **ddr_wck)
            self.des(o=self.out.dq_i[i], i=self.pads.dq[i],   name=f'dq_i{i}', **ddr_wck)

        # Output enable signals
        self.comb += [
            self.pads.dmi_oe.eq(delay(self.out.dmi_oe, cycles=Serializer.LATENCY)),
            self.pads.rdqs_oe.eq(delay(self.out.rdqs_oe, cycles=Serializer.LATENCY)),
            self.pads.dq_oe.eq(delay(self.out.dq_oe, cycles=Serializer.LATENCY)),
        ]



#
# This file is part of LiteDRAM.
#
# Copyright (c) 2015 Sebastien Bourdeauducq <sb@m-labs.hk>
# Copyright (c) 2016-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litedram.phy import dfi
from litex.soc.interconnect.csr import *

# PhaseInjector ------------------------------------------------------------------------------------

class PhaseInjector(Module, AutoCSR):
    def __init__(self, phase):
        self._command       = CSRStorage(fields=[
            CSRField("cs",   size=1, description="DFI chip select bus"),
            CSRField("we",   size=1, description="DFI write enable bus"),
            CSRField("cas",  size=1, description="DFI column address strobe bus"),
            CSRField("ras",  size=1, description="DFI row address strobe bus"),
            CSRField("wren", size=1, description="DFI write data enable bus"),
            CSRField("rden", size=1, description="DFI read data enable bus"),
        ], description="Control DFI signals on a single phase")

        self._command_issue = CSR() # description="The command gets commited on a write to this register"
        self._address       = CSRStorage(len(phase.address), reset_less=True,  description="DFI address bus")
        self._baddress      = CSRStorage(len(phase.bank),    reset_less=True,  description="DFI bank address bus")
        self._wrdata        = CSRStorage(len(phase.wrdata),  reset_less=True,  description="DFI write data bus")
        self._rddata        = CSRStatus(len(phase.rddata), description="DFI read data bus")

        # # #

        self.comb += [
            If(self._command_issue.re,
                phase.cs_n.eq(Replicate(~self._command.fields.cs, len(phase.cs_n))),
                phase.we_n.eq(~self._command.fields.we),
                phase.cas_n.eq(~self._command.fields.cas),
                phase.ras_n.eq(~self._command.fields.ras)
            ).Else(
                phase.cs_n.eq(Replicate(1, len(phase.cs_n))),
                phase.we_n.eq(1),
                phase.cas_n.eq(1),
                phase.ras_n.eq(1)
            ),
            phase.address.eq(self._address.storage),
            phase.bank.eq(self._baddress.storage),
            phase.wrdata_en.eq(self._command_issue.re & self._command.fields.wren),
            phase.rddata_en.eq(self._command_issue.re & self._command.fields.rden),
            phase.wrdata.eq(self._wrdata.storage),
            phase.wrdata_mask.eq(0)
        ]
        self.sync += If(phase.rddata_valid, self._rddata.status.eq(phase.rddata))

# DFIInjector --------------------------------------------------------------------------------------

class DFIInjector(Module, AutoCSR):
    def __init__(self, addressbits, bankbits, nranks, databits, nphases=1):
        self.slave   = dfi.Interface(addressbits, bankbits, nranks, databits, nphases)
        self.master  = dfi.Interface(addressbits, bankbits, nranks, databits, nphases)
        csr_dfi      = dfi.Interface(addressbits, bankbits, nranks, databits, nphases)

        self.ext_dfi     = dfi.Interface(addressbits, bankbits, nranks, databits, nphases)
        self.ext_dfi_sel = Signal()

        self._control = CSRStorage(fields=[
            CSRField("sel",     size=1, values=[
                ("``0b0``", "Software (CPU) control."),
                ("``0b1``", "Hardware control (default)."),
            ], reset=0b1), # Defaults to HW control.
            CSRField("cke",     size=1, description="DFI clock enable bus"),
            CSRField("odt",     size=1, description="DFI on-die termination bus"),
            CSRField("reset_n", size=1, description="DFI clock reset bus"),
        ], description="Control DFI signals common to all phases")

        for n, phase in enumerate(csr_dfi.phases):
            setattr(self.submodules, "pi" + str(n), PhaseInjector(phase))

        # # #

        self.comb += [
            # Hardware Control.
            # -----------------
            If(self._control.fields.sel,
                # Through External DFI.
                If(self.ext_dfi_sel,
                    self.ext_dfi.connect(self.master)
                # Through LiteDRAM controller.
                ).Else(
                    self.slave.connect(self.master)
                )
            # Software Control (through CSRs).
            # --------------------------------
            ).Else(
                csr_dfi.connect(self.master)
            )
        ]
        for i in range(nranks):
            self.comb += [phase.cke[i].eq(self._control.fields.cke) for phase in csr_dfi.phases]
            self.comb += [phase.odt[i].eq(self._control.fields.odt) for phase in csr_dfi.phases if hasattr(phase, "odt")]
        self.comb += [phase.reset_n.eq(self._control.fields.reset_n) for phase in csr_dfi.phases if hasattr(phase, "reset_n")]
